#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################

import pytest
import pyomo.environ as pyo
from pyomo.util.check_units import assert_units_consistent

from idaes.core import FlowsheetBlock
import idaes.core.util.scaling as iscale
from idaes.models.unit_models import HeatExchangerFlowPattern
from idaes.models.properties.modular_properties import GenericParameterBlock
from idaes.models_extra.power_generation.properties.natural_gas_PR import (
    get_prop,
    EosType,
)
from idaes.models_extra.power_generation.unit_models import (
    CrossFlowHeatExchanger1D,
    CrossFlowHeatExchanger1DInitializer,
)
import idaes.core.util.model_statistics as mstat
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.model_diagnostics import DiagnosticsToolbox

# Set up solver
optarg = {
    "constr_viol_tol": 1e-8,
    "nlp_scaling_method": "user-scaling",
    "linear_solver": "ma57",
    "OF_ma57_automatic_scaling": "yes",
    "max_iter": 350,
    "tol": 1e-8,
    "halt_on_ampl_error": "no",
}


def _create_model(pressure_drop):
    m = pyo.ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.h2_side_prop_params = GenericParameterBlock(
        **get_prop(["H2", "H2O", "Ar", "N2"], {"Vap"}, eos=EosType.IDEAL),
        doc="H2O + H2 gas property parameters",
    )
    m.fs.heat_exchanger = CrossFlowHeatExchanger1D(
        has_holdup=True,
        dynamic=False,
        cold_side={
            "property_package": m.fs.h2_side_prop_params,
            "has_holdup": False,
            "dynamic": False,
            "has_pressure_change": pressure_drop,
            "transformation_method": "dae.finite_difference",
            "transformation_scheme": "FORWARD",
        },
        hot_side={
            "property_package": m.fs.h2_side_prop_params,
            "has_holdup": False,
            "dynamic": False,
            "has_pressure_change": pressure_drop,
            "transformation_method": "dae.finite_difference",
            "transformation_scheme": "BACKWARD",
        },
        shell_is_hot=True,
        flow_type=HeatExchangerFlowPattern.countercurrent,
        finite_elements=12,
        tube_arrangement="staggered",
    )

    hx = m.fs.heat_exchanger

    hx.hot_side_inlet.flow_mol.fix(2619.7)
    hx.hot_side_inlet.temperature.fix(971.6)
    hx.hot_side_inlet.pressure.fix(1.2e5)
    hx.hot_side_inlet.mole_frac_comp[0, "H2"].fix(0.79715)
    hx.hot_side_inlet.mole_frac_comp[0, "H2O"].fix(0.20177)
    hx.hot_side_inlet.mole_frac_comp[0, "Ar"].fix(0.00086358)
    hx.hot_side_inlet.mole_frac_comp[0, "N2"].fix(0.00021589)

    hx.cold_side_inlet.flow_mol.fix(2619.7)
    hx.cold_side_inlet.temperature.fix(446.21)
    hx.cold_side_inlet.pressure.fix(1.2e5)
    hx.cold_side_inlet.mole_frac_comp[0, "H2"].fix(0.36203)
    hx.cold_side_inlet.mole_frac_comp[0, "H2O"].fix(0.63689)
    hx.cold_side_inlet.mole_frac_comp[0, "Ar"].fix(0.00086358)
    hx.cold_side_inlet.mole_frac_comp[0, "N2"].fix(0.00021589)

    hx.di_tube.fix(0.0525018)
    hx.thickness_tube.fix(0.0039116)
    hx.length_tube_seg.fix(4.3)
    hx.number_passes.fix(12)
    hx.number_columns_per_pass.fix(50)
    hx.number_rows_per_pass.fix(25)

    hx.pitch_x.fix(0.1)
    hx.pitch_y.fix(0.1)
    hx.therm_cond_wall = 43.0
    hx.rfouling_tube = 0.0001
    hx.rfouling_shell = 0.0001
    hx.fcorrection_htc_tube.fix(1)
    hx.fcorrection_htc_shell.fix(1)
    if pressure_drop:
        hx.fcorrection_dp_tube.fix(1)
        hx.fcorrection_dp_shell.fix(1)

    hx.cp_wall.value = 502.4

    pp = m.fs.h2_side_prop_params
    pp.set_default_scaling("enth_mol_phase", 1e-3)
    pp.set_default_scaling("pressure", 1e-5)
    pp.set_default_scaling("temperature", 1e-2)
    pp.set_default_scaling("flow_mol", 1e-3)

    _mf_scale = {
        "H2": 1,
        "H2O": 1,
        "N2": 10,
        "Ar": 10,
    }
    for comp, s in _mf_scale.items():
        pp.set_default_scaling("mole_frac_comp", s, index=comp)
        pp.set_default_scaling("mole_frac_phase_comp", s, index=("Vap", comp))
        pp.set_default_scaling("flow_mol_phase_comp", s * 1e-3, index=("Vap", comp))

    shell = hx.hot_side
    tube = hx.cold_side
    iscale.set_scaling_factor(shell.area, 1e-1)
    iscale.set_scaling_factor(shell.heat, 1e-6)
    iscale.set_scaling_factor(tube.area, 1)
    iscale.set_scaling_factor(tube.heat, 1e-6)
    iscale.set_scaling_factor(shell.enthalpy_flow_dx, 1e-7)
    iscale.set_scaling_factor(tube.enthalpy_flow_dx, 1e-7)
    iscale.set_scaling_factor(hx.heat_holdup, 1e-8)

    iscale.calculate_scaling_factors(m)

    return m


def _check_model_statistics(m, deltaP):
    fixed_unused_var_set = {
        "fs.h2_side_prop_params.H2.omega",
        "fs.h2_side_prop_params.H2.pressure_crit",
        "fs.h2_side_prop_params.H2.temperature_crit",
        "fs.h2_side_prop_params.H2.cp_mol_ig_comp_coeff_G",
        "fs.h2_side_prop_params.H2.cp_mol_ig_comp_coeff_H",
        "fs.h2_side_prop_params.H2O.omega",
        "fs.h2_side_prop_params.H2O.pressure_crit",
        "fs.h2_side_prop_params.H2O.temperature_crit",
        "fs.h2_side_prop_params.H2O.cp_mol_ig_comp_coeff_G",
        "fs.h2_side_prop_params.H2O.cp_mol_ig_comp_coeff_H",
        "fs.h2_side_prop_params.H2O.pressure_sat_comp_coeff_A",
        "fs.h2_side_prop_params.H2O.pressure_sat_comp_coeff_B",
        "fs.h2_side_prop_params.H2O.pressure_sat_comp_coeff_C",
        "fs.h2_side_prop_params.Ar.omega",
        "fs.h2_side_prop_params.Ar.pressure_crit",
        "fs.h2_side_prop_params.Ar.temperature_crit",
        "fs.h2_side_prop_params.Ar.cp_mol_ig_comp_coeff_G",
        "fs.h2_side_prop_params.Ar.cp_mol_ig_comp_coeff_H",
        "fs.h2_side_prop_params.N2.omega",
        "fs.h2_side_prop_params.N2.pressure_crit",
        "fs.h2_side_prop_params.N2.temperature_crit",
        "fs.h2_side_prop_params.N2.cp_mol_ig_comp_coeff_G",
        "fs.h2_side_prop_params.N2.cp_mol_ig_comp_coeff_H",
    }
    if not deltaP:
        fixed_unused_var_set.add("fs.heat_exchanger.delta_elevation")

    for var in mstat.fixed_unused_variables_set(m):
        assert var.name in fixed_unused_var_set

    unfixed_unused_var_set = {
        "fs.heat_exchanger.hot_side.material_flow_dx[0.0,0.0,Vap,H2]",
        "fs.heat_exchanger.hot_side.material_flow_dx[0.0,0.0,Vap,H2O]",
        "fs.heat_exchanger.hot_side.material_flow_dx[0.0,0.0,Vap,Ar]",
        "fs.heat_exchanger.hot_side.material_flow_dx[0.0,0.0,Vap,N2]",
        "fs.heat_exchanger.hot_side.enthalpy_flow_dx[0.0,0.0,Vap]",
        "fs.heat_exchanger.hot_side.pressure_dx[0.0,0.0]",
        "fs.heat_exchanger.cold_side.material_flow_dx[0.0,1.0,Vap,H2]",
        "fs.heat_exchanger.cold_side.material_flow_dx[0.0,1.0,Vap,H2O]",
        "fs.heat_exchanger.cold_side.material_flow_dx[0.0,1.0,Vap,Ar]",
        "fs.heat_exchanger.cold_side.material_flow_dx[0.0,1.0,Vap,N2]",
        "fs.heat_exchanger.cold_side.enthalpy_flow_dx[0.0,1.0,Vap]",
        "fs.heat_exchanger.cold_side.pressure_dx[0.0,1.0]",
    }

    for var in mstat.unused_variables_set(m) - mstat.fixed_unused_variables_set(m):
        assert var.name in unfixed_unused_var_set

    assert len(mstat.deactivated_constraints_set(m)) == 0


@pytest.fixture
def model_no_dP():
    m = _create_model(pressure_drop=False)
    return m


@pytest.mark.component
def test_initialization(model_no_dP):
    m = model_no_dP

    assert degrees_of_freedom(m) == 0
    _check_model_statistics(m, deltaP=False)

    initializer = m.fs.heat_exchanger.default_initializer(
        solver="ipopt", solver_options=optarg
    )
    assert isinstance(initializer, CrossFlowHeatExchanger1DInitializer)
    initializer.initialize(model=m.fs.heat_exchanger)

    assert degrees_of_freedom(m) == 0
    _check_model_statistics(m, deltaP=False)
    assert pyo.value(
        m.fs.heat_exchanger.hot_side_outlet.temperature[0]
    ) == pytest.approx(485.34, abs=1e-1)
    assert pyo.value(
        m.fs.heat_exchanger.cold_side_outlet.temperature[0]
    ) == pytest.approx(911.47, abs=1e-1)


@pytest.mark.integration
def test_structural_issues_no_dP(model_no_dP):
    dt = DiagnosticsToolbox(model_no_dP)
    dt.assert_no_structural_warnings(ignore_evaluation_errors=True)


@pytest.mark.integration
def test_numerical_issues_no_dP(model_no_dP):
    # Model will already be initialized if the component test is run,
    # but reinitialize in case integration tests are run alone
    initializer = model_no_dP.fs.heat_exchanger.default_initializer(
        solver="ipopt", solver_options=optarg
    )
    initializer.initialize(model=model_no_dP.fs.heat_exchanger)

    m_scaled = pyo.TransformationFactory("core.scale_model").create_using(
        model_no_dP, rename=False
    )

    dt = DiagnosticsToolbox(m_scaled)
    dt.assert_no_numerical_warnings()


@pytest.fixture
def model_dP():
    m = _create_model(pressure_drop=True)
    return m


@pytest.mark.component
def test_initialization_dP(model_dP):
    m = model_dP

    assert degrees_of_freedom(m) == 0
    _check_model_statistics(m, deltaP=True)

    initializer = m.fs.heat_exchanger.default_initializer(
        solver="ipopt", solver_options=optarg
    )
    assert isinstance(initializer, CrossFlowHeatExchanger1DInitializer)
    initializer.initialize(m.fs.heat_exchanger)

    assert degrees_of_freedom(m) == 0
    _check_model_statistics(m, deltaP=True)

    assert pyo.value(
        m.fs.heat_exchanger.hot_side_outlet.temperature[0]
    ) == pytest.approx(485.34, abs=1e-1)
    assert pyo.value(
        m.fs.heat_exchanger.cold_side_outlet.temperature[0]
    ) == pytest.approx(911.47, abs=1e-1)
    assert pyo.value(m.fs.heat_exchanger.hot_side_outlet.pressure[0]) == pytest.approx(
        118870.08569, abs=1
    )
    assert pyo.value(m.fs.heat_exchanger.cold_side_outlet.pressure[0]) == pytest.approx(
        111418.71399, abs=1
    )


@pytest.mark.integration
def test_structural_issues_dP(model_dP):
    dt = DiagnosticsToolbox(model_dP)
    dt.assert_no_structural_warnings(ignore_evaluation_errors=True)


@pytest.mark.integration
def test_numerical_issues_dP(model_dP):
    # Model will already be initialized if the component test is run,
    # but reinitialize in case integration tests are run alone
    initializer = model_dP.fs.heat_exchanger.default_initializer(
        solver="ipopt", solver_options=optarg
    )
    initializer.initialize(model=model_dP.fs.heat_exchanger)

    m_scaled = pyo.TransformationFactory("core.scale_model").create_using(
        model_dP, rename=False
    )

    dt = DiagnosticsToolbox(m_scaled)
    dt.assert_no_numerical_warnings()



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Methods for calculating pure component properties from:

The Properties of Gases & Liquids, 4th Edition
Reid, Prausnitz and Polling, 1987, McGraw-Hill

All parameter indices and units based on conventions used by the source
"""
# TODO: Missing doc strings
# pylint: disable=missing-class-docstring
# pylint: disable=missing-function-docstring

from pyomo.environ import exp, log, Var, units as pyunits

from idaes.core.util.misc import set_param_from_config


# -----------------------------------------------------------------------------
class RPP4(object):
    # -----------------------------------------------------------------------------
    # Heat capacities, enthalpies and entropies
    class cp_mol_ig_comp:
        @staticmethod
        def build_parameters(cobj):
            cobj.cp_mol_ig_comp_coeff_A = Var(
                doc="Parameter A for ideal gas molar heat capacity",
                units=pyunits.J / pyunits.mol / pyunits.K,
            )
            set_param_from_config(cobj, param="cp_mol_ig_comp_coeff", index="A")

            cobj.cp_mol_ig_comp_coeff_B = Var(
                doc="Parameter B for ideal gas molar heat capacity",
                units=pyunits.J / pyunits.mol / pyunits.K**2,
            )
            set_param_from_config(cobj, param="cp_mol_ig_comp_coeff", index="B")

            cobj.cp_mol_ig_comp_coeff_C = Var(
                doc="Parameter C for ideal gas molar heat capacity",
                units=pyunits.J / pyunits.mol / pyunits.K**3,
            )
            set_param_from_config(cobj, param="cp_mol_ig_comp_coeff", index="C")

            cobj.cp_mol_ig_comp_coeff_D = Var(
                doc="Parameter D for ideal gas molar heat capacity",
                units=pyunits.J / pyunits.mol / pyunits.K**4,
            )
            set_param_from_config(cobj, param="cp_mol_ig_comp_coeff", index="D")

        @staticmethod
        def return_expression(b, cobj, T):
            # Specific heat capacity
            T = pyunits.convert(T, to_units=pyunits.K)

            cp = (
                cobj.cp_mol_ig_comp_coeff_D * T**3
                + cobj.cp_mol_ig_comp_coeff_C * T**2
                + cobj.cp_mol_ig_comp_coeff_B * T
                + cobj.cp_mol_ig_comp_coeff_A
            )

            units = b.params.get_metadata().derived_units
            return pyunits.convert(cp, units.HEAT_CAPACITY_MOLE)

    class enth_mol_ig_comp:
        @staticmethod
        def build_parameters(cobj):
            if not hasattr(cobj, "cp_mol_ig_comp_coeff_A"):
                RPP4.cp_mol_ig_comp.build_parameters(cobj)

            if cobj.parent_block().config.include_enthalpy_of_formation:
                units = cobj.parent_block().get_metadata().derived_units

                cobj.enth_mol_form_vap_comp_ref = Var(
                    doc="Vapor phase molar heat of formation @ Tref",
                    units=units.ENERGY_MOLE,
                )
                set_param_from_config(cobj, param="enth_mol_form_vap_comp_ref")

        @staticmethod
        def return_expression(b, cobj, T):
            # Specific enthalpy
            T = pyunits.convert(T, to_units=pyunits.K)
            Tr = pyunits.convert(b.params.temperature_ref, to_units=pyunits.K)

            units = b.params.get_metadata().derived_units

            h_form = (
                cobj.enth_mol_form_vap_comp_ref
                if b.params.config.include_enthalpy_of_formation
                else 0 * units.ENERGY_MOLE
            )

            h = (
                pyunits.convert(
                    (cobj.cp_mol_ig_comp_coeff_D / 4) * (T**4 - Tr**4)
                    + (cobj.cp_mol_ig_comp_coeff_C / 3) * (T**3 - Tr**3)
                    + (cobj.cp_mol_ig_comp_coeff_B / 2) * (T**2 - Tr**2)
                    + cobj.cp_mol_ig_comp_coeff_A * (T - Tr),
                    units.ENERGY_MOLE,
                )
                + h_form
            )

            return h

    class entr_mol_ig_comp:
        @staticmethod
        def build_parameters(cobj):
            if not hasattr(cobj, "cp_mol_ig_comp_coeff_A"):
                RPP4.cp_mol_ig_comp.build_parameters(cobj)

            units = cobj.parent_block().get_metadata().derived_units

            cobj.entr_mol_form_vap_comp_ref = Var(
                doc="Vapor phase molar entropy of formation @ Tref",
                units=units.ENTROPY_MOLE,
            )
            set_param_from_config(cobj, param="entr_mol_form_vap_comp_ref")

        @staticmethod
        def return_expression(b, cobj, T):
            # Specific entropy
            T = pyunits.convert(T, to_units=pyunits.K)
            Tr = pyunits.convert(b.params.temperature_ref, to_units=pyunits.K)

            units = b.params.get_metadata().derived_units

            s = (
                pyunits.convert(
                    (cobj.cp_mol_ig_comp_coeff_D / 3) * (T**3 - Tr**3)
                    + (cobj.cp_mol_ig_comp_coeff_C / 2) * (T**2 - Tr**2)
                    + cobj.cp_mol_ig_comp_coeff_B * (T - Tr)
                    + cobj.cp_mol_ig_comp_coeff_A * log(T / Tr),
                    units.ENTROPY_MOLE,
                )
                + cobj.entr_mol_form_vap_comp_ref
            )

            return s

    # -----------------------------------------------------------------------------
    # Saturation pressure
    # Note that this equation in not valid beyond the critical temperature
    class pressure_sat_comp:
        @staticmethod
        def build_parameters(cobj):
            cobj.pressure_sat_comp_coeff_A = Var(
                doc="Coefficient A for calculating Psat", units=pyunits.dimensionless
            )
            set_param_from_config(cobj, param="pressure_sat_comp_coeff", index="A")

            cobj.pressure_sat_comp_coeff_B = Var(
                doc="Coefficient B for calculating Psat", units=pyunits.dimensionless
            )
            set_param_from_config(cobj, param="pressure_sat_comp_coeff", index="B")

            cobj.pressure_sat_comp_coeff_C = Var(
                doc="Coefficient C for calculating Psat", units=pyunits.dimensionless
            )
            set_param_from_config(cobj, param="pressure_sat_comp_coeff", index="C")

            cobj.pressure_sat_comp_coeff_D = Var(
                doc="Coefficient D for calculating Psat", units=pyunits.dimensionless
            )
            set_param_from_config(cobj, param="pressure_sat_comp_coeff", index="D")

        @staticmethod
        def return_expression(b, cobj, T, dT=False):
            if dT:
                return RPP4.pressure_sat_comp.dT_expression(b, cobj, T)

            x = 1 - T / cobj.temperature_crit

            return (
                exp(
                    (1 - x) ** -1
                    * (
                        cobj.pressure_sat_comp_coeff_A * x
                        + cobj.pressure_sat_comp_coeff_B * x**1.5
                        + cobj.pressure_sat_comp_coeff_C * x**3
                        + cobj.pressure_sat_comp_coeff_D * x**6
                    )
                )
                * cobj.pressure_crit
            )

        @staticmethod
        def dT_expression(b, cobj, T):
            x = 1 - T / cobj.temperature_crit

            return -RPP4.pressure_sat_comp.return_expression(b, cobj, T) * (
                (
                    cobj.pressure_sat_comp_coeff_A
                    + 1.5 * cobj.pressure_sat_comp_coeff_B * x**0.5
                    + 3 * cobj.pressure_sat_comp_coeff_C * x**2
                    + 6 * cobj.pressure_sat_comp_coeff_D * x**5
                )
                / T
                + (cobj.temperature_crit / T**2)
                * (
                    cobj.pressure_sat_comp_coeff_A * x
                    + cobj.pressure_sat_comp_coeff_B * x**1.5
                    + cobj.pressure_sat_comp_coeff_C * x**3
                    + cobj.pressure_sat_comp_coeff_D * x**6
                )
            )



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
import os
from copy import deepcopy

from .atom import Atom
from .canvas import Canvas
from .parsers.PDB import readPointsAndAtomsFromPDB, writeDesignToPDB
from .parsers.XYZ import readPointsAndAtomsFromXYZ, writeDesignToXYZ
from .parsers.CFG import readPointsAndAtomsFromCFG, writeDesignToCFG
from .parsers.POSCAR import readPointsAndAtomsFromPOSCAR, writeDesignToPOSCAR


class Design(object):
    """A class used to represent material designs.

    This class combines a ``Canvas`` objects and a list of contents.
    It assigns an element (possibly None) to each point in the ``Canvas``.
    This generally works for any type of content, but it is intended
    to work with ``Atom`` objects and can be used to generate CFG, PDB, POSCAR,
    and XYZ files.
    """

    # === STANDARD CONSTRUCTOR
    def __init__(self, Canvas_=None, Contents=None):
        """Initialize a Design object from standard data."""
        if Canvas_ is not None and Contents is None:
            Contents = [None] * len(Canvas_)
        elif Canvas_ is not None and isinstance(Contents, Atom):
            Contents = [Contents] * len(Canvas_)
        elif Canvas_ is None:
            Canvas_ = Canvas()
            Contents = []
        self._Canvas = Canvas_
        self._Contents = Contents

    # === CONSTRUCTOR - From PDB
    @classmethod
    def fromPDB(cls, filename, DefaultNN=0):
        """Make a Design by reading from PDB file.

        Args:
            filename(str): PDB file to read from.
            DefaultNN(int, optional): Optional, the default number of nearest
                neighbors to initialize the Canvas with.

        Returns:
            (Design) A new Design.

        """
        Pts, Atoms = readPointsAndAtomsFromPDB(filename)
        return cls(Canvas(Points=Pts, DefaultNN=DefaultNN), Atoms)

    # === CONSTRUCTOR - From XYZ
    @classmethod
    def fromXYZ(cls, filename, DefaultNN=0):
        """Make a Design by reading from XYZ file.

        Args:
            filename(str): XYZ file to read from.
            DefaultNN(int, optional): Optional, the default number of nearest
                neighbors to initialize the Canvas with.

        Returns:
            (Design) A new Design.

        """
        Pts, Atoms = readPointsAndAtomsFromXYZ(filename)
        return cls(Canvas(Points=Pts, DefaultNN=DefaultNN), Atoms)

    # === CONSTRUCTOR - From CFG
    @classmethod
    def fromCFG(cls, filename, DefaultNN=0):
        """Makes a Design by reading from CFG file.

        Args:
            filename(str): CFG file to read from.
            DefaultNN(int, optional): Optional, the default number of nearest
                neighbors to initialize the Canvas with.

        Returns:
            (Design) A new Design.

        """
        Pts, Atoms = readPointsAndAtomsFromCFG(filename)
        return cls(Canvas(Points=Pts, DefaultNN=DefaultNN), Atoms)

    # === CONSTRUCTOR - From POSCAR/CONTCAR
    @classmethod
    def fromPOSCAR(cls, filename, DefaultNN=0):
        """Make a Design by reading from POSCAR file.

        Args:
            filename(str): POSCAR file to read from.
            DefaultNN(int, optional): Optional, the default number of nearest
                neighbors to initialize the Canvas with.

        Returns:
            (Design) A new Design.

        """
        Pts, Atoms = readPointsAndAtomsFromPOSCAR(filename)
        return cls(Canvas(Points=Pts, DefaultNN=DefaultNN), Atoms)

    fromCONTCAR = fromPOSCAR
    """Makes a Design by reading from CONTCAR file.
       See documentation for fromPOSCAR. """

    # === MANIPULATION METHODS
    def setContent(self, i, Elem):
        """Set content at a particular location.

        Args:
            i(int): Index to set.
            Elem(Atom/Any): Element to place in Contents.

        Returns:
            None.

        """
        self._Contents[i] = Elem

    def setContents(self, Elem):
        """Set content across all locations.

        Args:
            Elem(Atom/Any): Element to place in each index of Contents.

        Returns:
            None.

        """
        for i in range(len(self.Contents)):
            self.setContent(i, Elem)

    def transform(self, TransF):
        """Transform the this Design according to a functor.

        Args:
            TransF(TransformFunc): Transformation to apply to Canvas.

        Returns:
            None.

        """
        self._Canvas.transform(TransF)

    def getTransformed(self, TransF):
        """Copy and transform this Design.

        Args:
            TransF(TransformFunc): Transformation to apply to Canvas.

        Returns:
            Design) Deepcopy of Design after transformation is applied.

        """
        result = deepcopy(self)
        result.transform(TransF)
        return result

    def add(self, P, Elem):
        """Add point and element to this Design.

        Args:
            P(numpy.ndarray): Point to add.
            Elem(Atom/Any): Content to add.

        Returns:

        """
        self._Canvas.addLocation(P)
        self._Contents.append(Elem)

    def addOther(self, other, blnAssertNotAlreadyInDesign=True):
        """Add another Design to this one.

        Args:
            other(Design): Design to append.
            blnAssertNotAlreadyInDesign(bool, optional): Optional, flag to enable
                assertion that all locations were new and unique.
                (Default value = True)

        Returns:
            None.

        """
        self._Canvas.addOther(other.Canvas, blnAssertNotAlreadyInDesign)
        self._Contents.extend(other.Contents)

    # === PROPERTY EVALUATION METHODS
    def __len__(self):
        """Get the size of this Canvas for this Design."""
        return len(self.Canvas)

    def __eq__(self, other):
        """Compare strict equality of two Designs."""
        return self.Contents == other.Contents and self.Canvas == other.Canvas

    def isEquivalentTo(self, other, blnPreserveIndexing=False, blnIgnoreVoid=True):
        """Compare equivilancy of two Designs.

        Args:
            blnIgnoreVoid:
            other(Design): other Design to compare against.
            blnPreserveIndexing(bool, optional): Optional, flag to determine if
                index order is considered. (Default value = False)
            blnIgnoreVoid(bool, optional): Optional, flag to determine if void
                (i.e., None or Atom() contents) should be considered.
                For example, if all solid atoms are equivalent, but there
                are void locations that do not match. (Default value = True)

        Returns:
            bool) True if Designs are considered equivalent.

        """
        if not blnIgnoreVoid and len(self) != len(other):
            return False
        if blnPreserveIndexing:
            return self == other
        for i, P in enumerate(self.Canvas.Points):
            if not other.Canvas.hasPoint(P):
                return False
            else:
                j = other.Canvas.getPointIndex(P)
                if self.Contents[i] != other.Contents[j]:
                    return False
        return True

    @property
    def NonVoidCount(self):
        """Count number of contents that are not considered void."""
        return sum(Elem is not None and Elem != Atom() for Elem in self.Contents)

    @property
    def NonVoidElems(self):
        """Get the set of non-void contents."""
        result = set(self.Contents)
        result.discard(None)
        result.discard(Atom())
        return result

    # === BASIC QUERY METHODS
    @property
    def Canvas(self):
        """Get the Canvas for this Design."""
        return self._Canvas

    @property
    def Contents(self):
        """Get the Contents for this Design."""
        return self._Contents

    # === REPORTING METHODS
    def toPDB(self, filename):
        """Write a Design to PDB file.

        Args:
            filename(str): PDB file to write to.

        Returns:
            None.

        """
        writeDesignToPDB(self, filename)

    def toXYZ(self, filename):
        """Write a Design to XYZ file.

        Args:
            filename(str): XYZ file to write to.

        Returns:
            None.

        """
        writeDesignToXYZ(self, filename)

    def toCFG(self, filename, GS=None, BBox=None, AuxPropMap=None, blnGroupByType=True):
        """Write a Design to CFG file.

        Args:
            blnGroupByType:
            AuxPropMap:
            BBox:
            filename(str): CFG file to write to.
            GS(float, optional): Optional, Global scaling to write to file
                (see CFG file format). (Default value = None)
            BBox(Parallelepiped, optional): Optional, Bounding box to write to file
                (see CFG file format). If not provided, calculates a
                rectangular prism 2x the necessary size to encompass points.
                (Default value = None)
            AuxPropMap(dict<tuple<str, optional): Optional, Auxiliary
                property map. Example: {('Energy','eV'):[0.0, 1.0, ... ]}
                (Default value = None)
            blnGroupByType(bool, optional): Optional, flag to group atoms by element.
                (see CFG file format). (Default value = True)

        Returns:
            None.

        """
        writeDesignToCFG(
            self,
            filename,
            GS=GS,
            BBox=BBox,
            AuxPropMap=AuxPropMap,
            blnGroupByType=blnGroupByType,
        )

    def toPOSCAR(
        self,
        filename,
        CommentLine=None,
        GS=None,
        BBox=None,
        Elems=None,
        blnUseDirect=True,
    ):
        """Write a Design to CFG file.

        Args:
            blnUseDirect:
            Elems:
            BBox:
            filename(str): CFG file to write to.
            CommentLine(str, optional): Optional, line to write at top of file. (Default value = None)
            GS(float, optional): Optional, Global scaling to write to file
                (see POSCAR file format). (Default value = None)
            BBox(Parallelepiped, optional): Optional, Bounding box to write to file
                (see POSCAR file format). If not provided, calculates a
                rectangular prism 2x the necessary size to encompass points.
                (Default value = None)
            Elems(list<Atom>, optional): Optional, order of elements to write to file.
                Only important because they are sometimes implicitly defined
                in other VASP files and may need to be in a definite order.
                (Default value = None)
            blnUseDirect(bool, optional): Optional, flag to switch between direct and
                cartesian flag of POSCAR file. (Default value = True)

        Returns:
            None.

        """
        writeDesignToPOSCAR(
            self,
            filename,
            CommentLine=CommentLine,
            GS=GS,
            BBox=BBox,
            Elems=Elems,
            blnUseDirect=blnUseDirect,
        )


def loadFromPDBs(filenames, folder=None):
    """Load a list of Designs from PDB files.

    Args:
        filenames (list<str>): List of files to read.
        folder (str): Optional, folder to prepend to filenames.
            (Default value = None)

    Returns:
        (list<Design>): List of Designs created.

    """
    result = []
    for filename in filenames:
        if folder is not None:
            filename = os.path.join(folder, filename)
        result.append(Design.fromPDB(filename))
    return result


def loadFromXYZs(filenames, folder=None):
    """Load a list of Designs from XYZ files.

    Args:
        filenames (list<str>): List of files to read.
        folder (str): Optional, folder to prepend to filenames.
            (Default value = None)

    Returns:
        (list<Design>): List of Designs created.

    """
    result = []
    for filename in filenames:
        if folder is not None:
            filename = os.path.join(folder, filename)
        result.append(Design.fromXYZ(filename))
    return result


def loadFromCFGs(filenames, folder=None):
    """Load a list of Designs from CFG files.

    Args:
        filenames (list<str>): List of files to read.
        folder (str): Optional, folder to prepend to filenames.
            (Default value = None)

    Returns:
        (list<Design>): List of Designs created.

    """
    result = []
    for filename in filenames:
        if folder is not None:
            filename = os.path.join(folder, filename)
        result.append(Design.fromCFG(filename))
    return result



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""Generate parameter and expression files for water
"""

__author__ = "John Eslick"

import os
import math
import pyomo.environ as pyo
from pyomo.common.fileutils import this_file_dir
from pyomo.common.fileutils import find_library
from idaes.core.util.math import smooth_max
from idaes.models.properties.general_helmholtz.helmholtz_parameters import (
    WriteParameters,
)


def thermal_conductivity_rule(m):
    """Thermal conductivity rule

    International Association for the Properties of Water and Steam (2011).
        IAPWS R15-11, "Release on the IAPWS Formulation 2011 for the
        Thermal Conductivity of Ordinary Water Substance,"
        URL: http://iapws.org/relguide/ThCond.pdf
    """

    L0 = {
        0: 2.443221e-3,
        1: 1.323095e-2,
        2: 6.770357e-3,
        3: -3.454586e-3,
        4: 4.096266e-4,
    }

    L1 = {
        (0, 0): 1.60397357,
        (1, 0): 2.33771842,
        (2, 0): 2.19650529,
        (3, 0): -1.21051378,
        (4, 0): -2.7203370,
        (0, 1): -0.646013523,
        (1, 1): -2.78843778,
        (2, 1): -4.54580785,
        (3, 1): 1.60812989,
        (4, 1): 4.57586331,
        (0, 2): 0.111443906,
        (1, 2): 1.53616167,
        (2, 2): 3.55777244,
        (3, 2): -0.621178141,
        (4, 2): -3.18369245,
        (0, 3): 0.102997357,
        (1, 3): -0.463045512,
        (2, 3): -1.40944978,
        (3, 3): 0.0716373224,
        (4, 3): 1.1168348,
        (0, 4): -0.0504123634,
        (1, 4): 0.0832827019,
        (2, 4): 0.275418278,
        (3, 4): 0.0,
        (4, 4): -0.19268305,
        (0, 5): 0.00609859258,
        (1, 5): -0.00719201245,
        (2, 5): -0.0205938816,
        (3, 5): 0.0,
        (4, 5): 0.012913842,
    }
    delta = m.delta
    tau = m.tau
    big_lam = 177.8514
    qdb = 1.0 / 0.40
    nu = 0.630
    gamma = 1.239
    xi0 = 0.13
    big_gam0 = 0.06
    Tbr = 1.5
    Tb = 1 / tau
    lib_path = find_library("general_helmholtz_external")
    m.cp = pyo.ExternalFunction(library=lib_path, function="cp")
    m.cv = pyo.ExternalFunction(library=lib_path, function="cv")
    m.mu = pyo.ExternalFunction(library=lib_path, function="mu")
    m.itc = pyo.ExternalFunction(
        library=lib_path, function="itc"
    )  # isothermal compressibility [1/MPa]
    deltchi = smooth_max(
        delta**2
        * (m.itc("h2o", delta, tau) - m.itc("h2o", delta, 1 / Tbr) * Tbr / Tb)
        * m.Pc
        / 1000,
        0,
        eps=1e-8,
    )
    xi = xi0 * (deltchi / big_gam0) ** (nu / gamma)
    y = qdb * xi
    kappa = m.cp("h2o", delta, tau) / m.cv("h2o", delta, tau)
    Z = (
        2.0
        / math.pi
        / y
        * (
            ((1 - 1.0 / kappa) * pyo.atan(y) + 1.0 / kappa * y)
            - (1 - pyo.exp(-1 / (1.0 / y + y**2 / 3 / delta**2)))
        )
    )
    cpb = m.cp("h2o", delta, tau) / m.R
    mub = m.mu("h2o", delta, tau)

    lam2 = big_lam * delta * Tb * cpb / mub * Z
    return lam2 + pyo.sqrt(1.0 / m.tau) / sum(
        L0val * m.tau**i for i, L0val in L0.items()
    ) * pyo.exp(
        m.delta
        * sum(
            (m.tau - 1) ** i * sum(L1[i, j] * (m.delta - 1) ** j for j in range(0, 6))
            for i in range(0, 5)
        )
    )


def viscosity_rule(mvisc):
    """Viscosity calculation for water

    International Association for the Properties of Water and Steam (2008).
        IAPWS R12-08, "Release on the IAPWS Formulation 2008 for the Viscosity
        of Ordinary Water Substance, URL: http://iapws.org/relguide/visc.pdf
    """

    H0 = {0: 1.67752, 1: 2.20462, 2: 0.6366564, 3: -0.241605}
    H1 = {
        (0, 0): 5.20094e-1,
        (1, 0): 8.50895e-2,
        (2, 0): -1.08374,
        (3, 0): -2.89555e-1,
        (4, 0): 0.0,
        (5, 0): 0.0,
        (0, 1): 2.22531e-1,
        (1, 1): 9.99115e-1,
        (2, 1): 1.88797,
        (3, 1): 1.26613,
        (4, 1): 0.0,
        (5, 1): 1.20573e-1,
        (0, 2): -2.81378e-1,
        (1, 2): -9.06851e-1,
        (2, 2): -7.72479e-1,
        (3, 2): -4.89837e-1,
        (4, 2): -2.57040e-1,
        (5, 2): 0.0,
        (0, 3): 1.61913e-1,
        (1, 3): 2.57399e-1,
        (2, 3): 0.0,
        (3, 3): 0.0,
        (4, 3): 0.0,
        (5, 3): 0.0,
        (0, 4): -3.25372e-2,
        (1, 4): 0.0,
        (2, 4): 0.0,
        (3, 4): 6.98452e-2,
        (4, 4): 0.0,
        (5, 4): 0.0,
        (0, 5): 0.0,
        (1, 5): 0.0,
        (2, 5): 0.0,
        (3, 5): 0.0,
        (4, 5): 8.72102e-3,
        (5, 5): 0.0,
        (0, 6): 0.0,
        (1, 6): 0.0,
        (2, 6): 0.0,
        (3, 6): -4.35673e-3,
        (4, 6): 0.0,
        (5, 6): -5.93264e-4,
    }

    return (
        1e2
        * pyo.sqrt(1.0 / mvisc.tau)
        / sum(H0val * mvisc.tau**i for i, H0val in H0.items())
        * pyo.exp(
            mvisc.delta
            * sum(
                (mvisc.tau - 1) ** i
                * sum(H1[i, j] * (mvisc.delta - 1) ** j for j in range(0, 7))
                for i in range(0, 6)
            )
        )
    )


def main(dry_run=False):
    """Generate parameter and expression files.

    Args:
        dry_run (bool): If dry run don't generate files

    Returns:
        WriteParameters
    """
    main_param_file = os.path.join(this_file_dir(), "h2o.json")
    we = WriteParameters(parameters=main_param_file)
    we.add(
        {
            "viscosity": viscosity_rule,
            "thermal_conductivity": thermal_conductivity_rule,
        }
    )
    we.write(dry_run=dry_run)
    return we


if __name__ == "__main__":
    main()



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
import logging
import os
from pathlib import Path
from shutil import rmtree


def create_module_scratch(module_name):
    """Create a scratch that is easily found (by developers) and removed, and doesn't
    clutter up the working directories.
    """
    path = Path("~").expanduser()
    for path_part in ".idaes", "_scratch", module_name:
        path = path / path_part
        if not path.exists():
            path.mkdir()
    return path


def rmtree_scratch(scratch_path):
    """Do our level best to remove all the temporary files."""
    try:
        rmtree(scratch_path, ignore_errors=True)
    except Exception:  # pylint: disable=W0703
        pass



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
from copy import deepcopy
from math import sqrt

import numpy as np

from .unit_cell_lattice import UnitCell, UnitCellLattice
from ..geometry import Cube
from ..tiling import CubicTiling
from ..transform_func import ScaleFunc, RotateFunc
from ...util.util import ListHasPoint


class FCCLattice(UnitCellLattice):
    RefIAD = sqrt(2) / 2

    # === STANDARD CONSTRUCTOR
    def __init__(self, IAD):
        RefUnitCellShape = Cube(1, BotBackLeftCorner=np.array([0, 0, 0], dtype=float))
        RefUnitCellTiling = CubicTiling(RefUnitCellShape)
        RefFracPositions = [
            np.array([0.0, 0.0, 0.0]),
            np.array([0.5, 0.5, 0.0]),
            np.array([0.0, 0.5, 0.5]),
            np.array([0.5, 0.0, 0.5]),
        ]
        RefUnitCell = UnitCell(RefUnitCellTiling, RefFracPositions)
        UnitCellLattice.__init__(self, RefUnitCell)
        self._IAD = FCCLattice.RefIAD  # IAD is set correctly after calling applyTransF
        self.applyTransF(ScaleFunc(IAD / FCCLattice.RefIAD))
        self._NthNeighbors = [
            [
                np.array([0.0, -0.5, 0.5]),
                np.array([-0.5, -0.5, 0.0]),
                np.array([-0.5, 0.0, 0.5]),
                np.array([0.5, -0.5, 0.0]),
                np.array([0.0, -0.5, -0.5]),
                np.array([-0.5, 0.0, -0.5]),
                np.array([-0.5, 0.5, 0.0]),
                np.array([0.0, 0.5, 0.5]),
                np.array([0.5, 0.0, 0.5]),
                np.array([0.5, 0.0, -0.5]),
                np.array([0.0, 0.5, -0.5]),
                np.array([0.5, 0.5, 0.0]),
            ]
        ]

    # === CONSTRUCTOR - Aligned with FCC {100}
    @classmethod
    def alignedWith100(cls, IAD):
        return cls(IAD)  # Default implementation

    # === CONSTRUCTOR - Aligned with FCC {111}
    @classmethod
    def alignedWith111(cls, IAD, blnTrianglesAlignedWithX=True):
        result = cls(IAD)
        thetaX = -np.pi * 0.25
        thetaY = -np.arctan2(-sqrt(2), 2)
        thetaZ = np.pi * 0.5 if blnTrianglesAlignedWithX else 0
        result.applyTransF(RotateFunc.fromXYZAngles(thetaX, thetaY, thetaZ))
        return result

    # === MANIPULATION METHODS
    def applyTransF(self, TransF):
        if isinstance(TransF, ScaleFunc):
            if TransF.isIsometric:
                self._IAD *= TransF.Scale[0]
            else:
                raise ValueError("FCCLattice applyTransF: Can only scale isometrically")
        UnitCellLattice.applyTransF(self, TransF)

    # === PROPERTY EVALUATION METHODS
    # NOTE: inherited from UnitCellLattice
    # def isOnLattice(self,P):

    def areNeighbors(self, P1, P2):
        return np.linalg.norm(P2 - P1) <= self.IAD

    def getNeighbors(self, P, layer=1):
        RefP = self._getConvertToReference(P)
        if layer > len(self._NthNeighbors):
            self._calculateNeighbors(layer)
        NBs = deepcopy(self._NthNeighbors[layer - 1])
        for NeighP in NBs:
            NeighP += RefP
            self._convertFromReference(NeighP)
        return NBs

    def _calculateNeighbors(self, layer):
        NList = [np.array([0, 0, 0], dtype=float)]
        for nb in self._NthNeighbors:
            NList.extend(nb)
        for _ in range(layer - len(self._NthNeighbors)):
            tmp = []
            for P in self._NthNeighbors[len(self._NthNeighbors) - 1]:
                for Q in self._NthNeighbors[0]:
                    N = P + Q
                    if not ListHasPoint(NList, N, 0.001 * FCCLattice.RefIAD):
                        tmp.append(N)
                        NList.append(N)
            self._NthNeighbors.append(tmp)

    # === BASIC QUERY METHODS
    @property
    def IAD(self):
        return self._IAD

    @property
    def FCC111LayerSpacing(self):
        return self.IAD * sqrt(2) / sqrt(3)

    @property
    def FCC100LayerSpacing(self):
        return self.IAD * sqrt(2) / 2

    @property
    def FCC110LayerSpacing(self):
        return self.IAD



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Gas viscosity at low pressures from Chapman Enskog theory using Lennard Jones parameters.
Note that LJ parameters are underdetermined when estimated from viscosity data (see
The Indeterminacy of the Values of Potential Parameters as Derived from Transport and
Virial Coefficient by Reichenberg D., 1973 for more information) so it's important to
use LJ parameters from the same source.
"""
from pyomo.environ import Var, units as pyunits
import pyomo.environ as pyo

from idaes.core.util.misc import set_param_from_config


class ChapmanEnskogLennardJones(object):
    """Class implementing physical properties from Chapman Enskog theory"""

    @staticmethod
    def build_lennard_jones_parameters(cobj):
        """Method to build Lennard Jones parameters"""
        units = cobj.parent_block().get_metadata().derived_units
        if not hasattr(cobj, "lennard_jones_sigma"):
            cobj.lennard_jones_sigma = Var(
                doc="Parameter sigma from Lennard Jones potential",
                units=units["length"],
            )
            set_param_from_config(cobj, param="lennard_jones_sigma")

        if not hasattr(cobj, "lennard_jones_epsilon_reduced"):
            cobj.lennard_jones_epsilon_reduced = Var(
                doc="Parameter epsilon from Lennard Jones potential reduced by Boltzmann constant",
                units=units["temperature"],
            )
            set_param_from_config(cobj, param="lennard_jones_epsilon_reduced")

    class visc_d_phase_comp(object):
        """Implementation of pure component dynamic viscosity from Chapman Enskog Theory"""

        @staticmethod
        def build_parameters(cobj, p):
            """Build Lennard Jones parameters and add callback for viscosity collision integral"""
            ChapmanEnskogLennardJones.build_lennard_jones_parameters(cobj)
            if not hasattr(cobj, "viscosity_collision_integral_callback"):
                cobj.viscosity_collision_integral_callback = (
                    collision_integral_neufeld_callback
                )

        @staticmethod
        def return_expression(b, cobj, p, T):
            """Return expression for visc_d_phase_comp"""
            # Properties of Gases and Liquids, Eq. 9.3.9
            units = b.params.get_metadata().derived_units

            T = pyunits.convert(T, to_units=pyunits.K)
            sigma = pyunits.convert(cobj.lennard_jones_sigma, pyunits.angstrom)
            M = pyunits.convert(cobj.mw, pyunits.g / pyunits.mol)
            T_dim = T / pyunits.convert(
                cobj.lennard_jones_epsilon_reduced, to_units=pyunits.K
            )
            Omega = cobj.viscosity_collision_integral_callback(T_dim)

            C = (
                26.69
                * pyunits.micropoise
                * pyunits.angstrom**2
                / pyo.sqrt(pyunits.g / pyunits.mol * pyunits.K)
            )
            visc = C * pyo.sqrt(M * T) / (sigma**2 * Omega)

            return pyunits.convert(visc, units["dynamic_viscosity"])


def collision_integral_kim_ross_callback(T_dim):
    """Equation 9.4.5 from Properties of Gases and Liquids.

    .. math::

      \\Omega(T^*) =1.604/\\sqrt{T^*}

    """
    return 1.604 / pyo.sqrt(T_dim)


def collision_integral_neufeld_callback(T_dim):
    """Equation 9.4.3 from Properties of Gases and Liquids.

    .. math::

      \\Omega(T^*) = A(T^*)^{-B} +  C\\exp(-DT^*) + E\\exp(-FT^*)

    in which :math:`A=1.16145`, :math:`B=0.14874`, :math:`C=0.52487`, :math:`D=0.77320`, :math:`E=2.16178`, and :math:`F=2.43787`.
    """
    A = 1.16145
    B = 0.14874
    C = 0.52487
    D = 0.77320
    E = 2.16178
    F = 2.43787
    return A * T_dim**-B + C * pyo.exp(-D * T_dim) + E * pyo.exp(-F * T_dim)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
from itertools import zip_longest
from types import ModuleType

from pyomo.common.dependencies import attempt_import
from pyomo.common.config import ConfigDict, ConfigValue
import pyomo.environ as pyo
from idaes.apps.grid_integration.utils import convert_marginal_costs_to_actual_costs

prescient, prescient_avail = attempt_import("prescient")


class PrescientPluginModule(ModuleType):
    def __init__(self, get_configuration, register_plugins):
        self.get_configuration = get_configuration
        self.register_plugins = register_plugins


class DoubleLoopCoordinator:
    """
    Coordinate Prescient, tracker and bidder.
    """

    def __init__(self, bidder, tracker, projection_tracker):
        """
        Initializes the DoubleLoopCoordinator object and registers functionalities
        in Prescient's plugin system.

        Arguments:
            bidder: an initialized bidder object

            tracker: an initialized tracker object

            projection_tracker: an initialized tracker object, this object is
                                mimicking the behaviror of the projection SCED in
                                Prescient and to projecting the system states
                                and updating bidder model.

        Returns:
            None
        """

        self.bidder = bidder
        self.tracker = tracker
        self.projection_tracker = projection_tracker

    def register_plugins(self, context, options, plugin_config):
        """
        Register functionalities in Prescient's plugin system.

        Arguments:
            context: Prescient plugin PluginRegistrationContext from prescient.plugins.plugin_registration

            options: Prescient options from prescient.simulator.config

            plugin_config: Prescient plugin config

        Returns:
            None
        """

        self.plugin_config = plugin_config

        context.register_initialization_callback(self.initialize_customized_results)
        context.register_for_hourly_stats(self.push_hourly_stats_to_forecaster)
        context.register_after_get_initial_actuals_model_for_sced_callback(
            self.update_static_params
        )
        context.register_after_get_initial_actuals_model_for_simulation_actuals_callback(
            self.update_static_params
        )
        context.register_after_get_initial_forecast_model_for_ruc_callback(
            self.update_static_params
        )
        context.register_before_ruc_solve_callback(self.bid_into_DAM)
        context.register_after_ruc_generation_callback(self.fetch_DA_prices)
        context.register_after_ruc_generation_callback(self.fetch_DA_dispatches)
        context.register_after_ruc_generation_callback(
            self.push_day_ahead_stats_to_forecaster
        )
        context.register_before_operations_solve_callback(self.bid_into_RTM)
        context.register_after_operations_callback(self.track_sced_signal)
        context.register_update_operations_stats_callback(self.update_observed_dispatch)
        context.register_after_ruc_activation_callback(self.activate_pending_DA_data)
        context.register_finalization_callback(self.write_plugin_results)

        return

    def get_configuration(self, key):
        """
        Register customized commandline options.

        Arguments:
            key: plugin name

        Returns:
            config: Prescient config dict
        """

        config = ConfigDict()

        # Add command line options
        config.declare(
            "bidding_generator",
            ConfigValue(
                domain=str,
                description="Specifies the generator we derive bidding strategies for.",
                default=None,
            ),
        ).declare_as_argument("--bidding-generator")

        return config

    @property
    def prescient_plugin_module(self):
        return PrescientPluginModule(self.get_configuration, self.register_plugins)

    def initialize_customized_results(self, options, simulator):
        """
        This method is outdated.
        """

        simulator.data_manager.extensions["customized_results"] = {}
        customized_results = simulator.data_manager.extensions["customized_results"]

        customized_results["Generator"] = []
        customized_results["Date"] = []
        customized_results["Hour"] = []
        customized_results["State"] = []
        customized_results["RUC Schedule"] = []
        customized_results["SCED Schedule"] = []
        customized_results["Power Output"] = []

        return

    def push_hourly_stats_to_forecaster(self, prescient_hourly_stats):
        """
        This method pushes the hourly stats from Prescient to the price forecaster
        once the hourly stats are published.

        Arguments:
            prescient_hourly_stats: Prescient HourlyStats object.

        Returns:
            None
        """

        self.bidder.forecaster.fetch_hourly_stats_from_prescient(prescient_hourly_stats)

    def push_day_ahead_stats_to_forecaster(
        self, options, simulator, day_ahead_result, uc_date, uc_hour
    ):
        """
        This method pushes the day-ahead market to the price forecaster after the
        UC is solved.

        Arguments:
            options: Prescient options from prescient.simulator.config.

            simulator: Prescient simulator.

            day_ahead_result: a Prescient RucPlan object.

            ruc_date: the date of the day-ahead market we bid into.

            ruc_hour: the hour the RUC is being solved in the day before.

        Returns:
            None
        """

        self.bidder.forecaster.fetch_day_ahead_stats_from_prescient(
            uc_date, uc_hour, day_ahead_result
        )

    def _update_bids(self, gen_dict, bids, start_hour, horizon):
        """
        This method takes bids and pass the information in the bids to generator
        dictionary from Prescient.

        Arguments:
            gen_dict: a dictionary from Prescient's RUC/SCED instance that stores
            the generator parameters.

            bids: the bids we want to pass into the market.

            start_hour: the effective start hour of the bid

            horizon: the length of the bids

        Returns:
            None
        """

        gen_name = self.bidder.bidding_model_object.model_data.gen_name

        def _update_p_cost(gen_dict, bids, param_name, start_hour, horizon):

            # update the "p_cost" element in the generator's dict
            gen_dict["p_cost"] = {
                "data_type": "time_series",
                "values": [
                    {
                        "data_type": "cost_curve",
                        "cost_curve_type": "piecewise",
                        "values": bids[t][gen_name]["p_cost"],
                    }
                    for t in range(start_hour, horizon + start_hour)
                ],
            }

            # because the p_cost is updated, so delete p_fuel
            if "p_fuel" in gen_dict:
                gen_dict.pop("p_fuel")

            return

        def _update_time_series_params(gen_dict, bids, param_name, start_hour, horizon):

            value_list = [
                bids[t][gen_name].get(param_name, None)
                for t in range(start_hour, start_hour + horizon)
            ]
            if param_name in gen_dict:
                gen_dict[param_name] = {
                    "data_type": "time_series",
                    "values": value_list,
                }

            return

        def _update_non_time_series_params(
            gen_dict, bids, param_name, start_hour, horizon
        ):
            if param_name in gen_dict:
                gen_dict[param_name] = bids[start_hour][gen_name].get(param_name, None)

            return

        param_update_func_map = {
            "p_cost": _update_p_cost,
            "p_max": _update_time_series_params,
            "p_min": _update_time_series_params,
            "p_min_agc": _update_time_series_params,
            "p_max_agc": _update_time_series_params,
            "fixed_commitment": _update_time_series_params,
            "min_up_time": _update_non_time_series_params,
            "min_down_time": _update_non_time_series_params,
            "startup_capacity": _update_time_series_params,
            "shutdown_capacity": _update_time_series_params,
            "startup_fuel": _update_non_time_series_params,
            "startup_cost": _update_non_time_series_params,
        }

        for param_name in bids[start_hour][gen_name].keys():
            update_func = param_update_func_map[param_name]
            update_func(gen_dict, bids, param_name, start_hour, horizon)

        return

    def _pass_DA_bid_to_prescient(self, options, ruc_instance, bids):
        """
        This method passes the bids into the RUC model for day-ahead market clearing.

        Arguments:
            options: Prescient options from prescient.simulator.config.

            ruc_instance: Prescient RUC object

            bids: the bids we want to pass into the day-ahead market.

        Returns:
            None
        """

        gen_name = self.bidder.bidding_model_object.model_data.gen_name

        # fetch the generator's parameter dictionary from Prescient UC instance
        gen_dict = ruc_instance.data["elements"]["generator"][gen_name]

        self._update_bids(gen_dict, bids, start_hour=0, horizon=options.ruc_horizon)

        return

    def assemble_project_tracking_signal(self, options, simulator, hour):
        """
        This function assembles the signals for the tracking model to estimate the
        state of the bidding model at the beginning of next RUC.

        Arguments:
            options: Prescient options from prescient.simulator.config.

            simulator: Prescient simulator.

            hour: the simulation hour.

        Returns:
            market_signals: the market signals to be tracked.
        """

        tracking_horizon = len(self.projection_tracker.time_set)

        market_signals = self._assemble_sced_tracking_market_signals(
            hour=hour,
            sced_dispatch=None,
            tracking_horizon=tracking_horizon,
        )
        return market_signals

    def project_tracking_trajectory(self, options, simulator, ruc_hour):
        """
        This function projects the full power dispatch trajectory from the
        tracking model so we can use it to update the bidding model, i.e. advance
        the time for the bidding model.

        Arguments:
            options: Prescient options from prescient.simulator.config.

            simulator: Prescient simulator.

            ruc_hour: the hour RUC is being solved

        Returns:
            full_projected_trajectory: the full projected power dispatch trajectory.
        """

        current_date = simulator.time_manager.current_time.date
        current_hour = simulator.time_manager.current_time.hour

        self._clone_tracking_model()

        for hour in range(ruc_hour, 24):

            # assemble market_signals
            market_signals = self.assemble_project_tracking_signal(
                options=options, simulator=simulator, hour=hour
            )
            # solve tracking
            self.projection_tracker.track_market_dispatch(
                market_dispatch=market_signals, date=current_date, hour=current_hour
            )

        # merge the trajectory
        full_projected_trajectory = {}
        for stat in self.tracker.daily_stats:
            full_projected_trajectory[stat] = self.tracker.daily_stats.get(
                stat
            ) + self.projection_tracker.daily_stats.get(stat)

        # clear the projection stats
        self.projection_tracker.daily_stats = None

        return full_projected_trajectory

    def _clone_tracking_model(self):
        """
        Clone the model in tracker and replace that of projection tracker. In this
        way, tracker and projection tracker have the same states before projection.

        Arguments:
            None

        Returns:
            None
        """

        # iterate all the variables and params and clone the values
        objects_list = [pyo.Var, pyo.Param]
        for obj in objects_list:
            for tracker_obj, proj_tracker_obj in zip_longest(
                self.tracker.model.component_objects(
                    obj, sort=pyo.SortComponents.alphabetizeComponentAndIndex
                ),
                self.projection_tracker.model.component_objects(
                    obj, sort=pyo.SortComponents.alphabetizeComponentAndIndex
                ),
            ):
                if tracker_obj.name != proj_tracker_obj.name:
                    raise ValueError(
                        f"Trying to copy the value of {tracker_obj} to {proj_tracker_obj}, but they do not have the same name and possibly not the corresponding objects. Please make sure tracker and projection tracker do not diverge. "
                    )
                for idx in tracker_obj.index_set():
                    if pyo.value(proj_tracker_obj[idx]) != pyo.value(tracker_obj[idx]):
                        proj_tracker_obj[idx] = round(pyo.value(tracker_obj[idx]), 4)

        return

    def _update_static_params(self, gen_dict):
        """
        Update static parameters in the Prescient generator parameter data dictionary depending on generator type.

        For a thermal generator, the p_cost data will be via ( MWh, $ ) pairs.
        For a renewable generator, the p_cost is a single cost.

        Args:
            gen_dict: Prescient generator parameter data dictionary.

        Returns:
            None
        """

        is_thermal = (
            self.bidder.bidding_model_object.model_data.generator_type == "thermal"
        )
        is_renewable = (
            self.bidder.bidding_model_object.model_data.generator_type == "renewable"
        )

        for param, value in self.bidder.bidding_model_object.model_data:
            if param == "gen_name" or value is None:
                continue
            elif (
                param in gen_dict
                and isinstance(gen_dict[param], dict)
                and gen_dict[param]["data_type"] == "time_series"
            ):
                # don't touch time varying things;
                # presumably they be updated later
                continue
            elif param == "p_cost":
                if is_thermal:
                    curve_value = convert_marginal_costs_to_actual_costs(value)
                    gen_dict[param] = {
                        "data_type": "cost_curve",
                        "cost_curve_type": "piecewise",
                        "values": curve_value,
                    }
                elif is_renewable:
                    gen_dict[param] = value
                else:
                    raise NotImplementedError(
                        "generator_type must be either 'thermal' or 'renewable'"
                    )
            else:
                gen_dict[param] = value

    def update_static_params(self, options, instance):
        """
        This method pass static generator parameters in Prescient
        immediately after a model is created.

        Arguments:
            options: Prescient options from prescient.simulator.config.

            instance: Prescient RUC object or SCED object.

        Returns:
            None
        """

        gen_name = self.bidder.bidding_model_object.model_data.gen_name
        gen_dict = instance.data["elements"]["generator"][gen_name]
        self._update_static_params(gen_dict)

        return

    def bid_into_DAM(self, options, simulator, ruc_instance, ruc_date, ruc_hour):
        """
        This function uses the bidding objects to bid into the day-ahead market
        (DAM).

        Arguments:
            options: Prescient options from prescient.simulator.config.

            simulator: Prescient simulator.

            ruc_instance: Prescient RUC object.

            ruc_date: the date of the day-ahead market we bid into.

            ruc_hour: the hour the RUC is being solved in the day before.

        Returns:
            None
        """

        # check if it is first day
        is_first_day = simulator.time_manager.current_time is None

        if not is_first_day:

            # solve rolling horizon to get the trajectory
            full_projected_trajectory = self.project_tracking_trajectory(
                options, simulator, options.ruc_execution_hour
            )
            # update the bidding model
            self.bidder.update_day_ahead_model(**full_projected_trajectory)

        # generate bids
        bids = self.bidder.compute_day_ahead_bids(date=ruc_date)

        if is_first_day:
            self.current_bids = bids
        self.next_bids = bids

        # pass to prescient
        self._pass_DA_bid_to_prescient(options, ruc_instance, bids)

        return

    def fetch_DA_prices(self, options, simulator, result, uc_date, uc_hour):
        """
        This method fetches the day-ahead market prices from unit commitment results,
        and save it as a coordinator attribute.

        Arguments:
            options: Prescient options from prescient.simulator.config.

            simulator: Prescient simulator.

            result: a Prescient RucPlan object.

            ruc_date: the date of the day-ahead market we bid into.

            ruc_hour: the hour the RUC is being solved in the day before.

        Returns:
            None
        """

        current_bus = self.bidder.bidding_model_object.model_data.bus
        is_first_day = simulator.time_manager.current_time is None

        DA_prices = [
            result.ruc_market.day_ahead_prices.get((current_bus, t)) for t in range(24)
        ]

        if is_first_day:
            self.current_avail_DA_prices = DA_prices
        else:
            self.current_avail_DA_prices = self.current_DA_prices + DA_prices
        self.next_DA_prices = DA_prices

    def fetch_DA_dispatches(self, options, simulator, result, uc_date, uc_hour):
        """
        This method fetches the day-ahead dispatches from unit commitment results,
        and save it as a coordinator attribute.

        Arguments:
            options: Prescient options from prescient.simulator.config.

            simulator: Prescient simulator.

            result: a Prescient RucPlan object.

            ruc_date: the date of the day-ahead market we bid into.

            ruc_hour: the hour the RUC is being solved in the day before.

        Returns:
            None
        """

        is_first_day = simulator.time_manager.current_time is None
        gen_name = self.bidder.bidding_model_object.model_data.gen_name
        gen_type = self.bidder.bidding_model_object.model_data.generator_type

        gen_type_dispatch_mapping = {
            "thermal": result.ruc_market.thermal_gen_cleared_DA,
            "renewable": result.ruc_market.renewable_gen_cleared_DA,
            "virtual": result.ruc_market.virtual_gen_cleared_DA,
        }

        dispatch_dict = gen_type_dispatch_mapping[gen_type]
        DA_dispatches = [dispatch_dict.get((gen_name, t)) for t in range(24)]

        if is_first_day:
            # self.current_DA_dispatches = DA_dispatches
            self.current_avail_DA_dispatches = DA_dispatches
        else:
            self.current_avail_DA_dispatches = (
                self.current_DA_dispatches + DA_dispatches
            )
        self.next_DA_dispatches = DA_dispatches

    def _pass_RT_bid_to_prescient(self, options, simulator, sced_instance, bids, hour):
        """
        This method passes the bids into the SCED model for real-time market
        clearing.

        Arguments:
            options: Prescient options from prescient.simulator.config.

            simulator: Prescient simulator.

            sced_instance: Prescient SCED object

            bids: the bids we want to pass into the real-time market.

            hour: the hour of the real-time market.

        Returns:
            None
        """

        gen_name = self.bidder.bidding_model_object.model_data.gen_name

        # fetch generator's parameter dictionary from SCED instance
        gen_dict = sced_instance.data["elements"]["generator"][gen_name]

        self._update_bids(gen_dict, bids, start_hour=hour, horizon=options.sced_horizon)

        return

    def bid_into_RTM(self, options, simulator, sced_instance):
        """
        This function bids into the real-time market. At this moment I just copy the
        corresponding day-ahead bid here.

        Arguments:
            options: Prescient options from prescient.simulator.config.

            simulator: Prescient simulator.

            sced_instance: Prescient SCED object.

        Returns:
            None
        """

        # fetch the bids
        date = simulator.time_manager.current_time.date
        hour = simulator.time_manager.current_time.hour

        bids = self.bidder.compute_real_time_bids(
            date=date,
            hour=hour,
            realized_day_ahead_prices=self.current_avail_DA_prices,
            realized_day_ahead_dispatches=self.current_avail_DA_dispatches,
        )

        # pass bids into sced model
        self._pass_RT_bid_to_prescient(options, simulator, sced_instance, bids, hour)

        return

    def assemble_sced_tracking_market_signals(
        self, options, simulator, sced_instance, hour
    ):
        """
        This function assembles the signals for the tracking model.

        Arguments:
            options: Prescient options from prescient.simulator.config.

            simulator: Prescient simulator.

            sced_instance: Prescient SCED object

            hour: the simulation hour.

        Returns:
            market_signals: the market signals to be tracked.
        """

        gen_name = self.bidder.bidding_model_object.model_data.gen_name

        # fecth the sced signals for the generation from sced instance
        sced_dispatch = sced_instance.data["elements"]["generator"][gen_name]["pg"][
            "values"
        ]
        tracking_horizon = len(self.tracker.time_set)

        market_signals = self._assemble_sced_tracking_market_signals(
            hour=hour,
            sced_dispatch=sced_dispatch,
            tracking_horizon=tracking_horizon,
        )

        return market_signals

    def _assemble_sced_tracking_market_signals(
        self, hour, sced_dispatch, tracking_horizon
    ):
        """
        This function assembles the signals for the tracking model.

        Arguments:

            hour: the simulation hour

            sced_dispatch: current sced dispatch (a list)

            tracking_horizon: length of the tracking horizon


        Returns:
            market_signals: the market signals to be tracked.
        """

        market_signals = []

        # append the sced dispatch
        if sced_dispatch is None:
            dispatch = self.current_DA_dispatches[hour]
        else:
            dispatch = sced_dispatch[0]
        market_signals.append(dispatch)

        # append corresponding RUC dispatch
        for t in range(hour + 1, hour + tracking_horizon):
            if t < len(self.current_avail_DA_dispatches):
                market_signals.append(self.current_avail_DA_dispatches[t])

        return market_signals

    def track_sced_signal(self, options, simulator, sced_instance, lmp_sced):
        """
        This methods uses the tracking object to track the current real-time market
        signals.

        Arguments:
            options: Prescient options from prescient.simulator.config.

            simulator: Prescient simulator.

            sced_instance: Prescient SCED object

            lmp_sced: Prescient SCED LMP object

        Returns:
            None

        """

        current_date = simulator.time_manager.current_time.date
        current_hour = simulator.time_manager.current_time.hour

        # get market signals
        market_signals = self.assemble_sced_tracking_market_signals(
            options=options,
            simulator=simulator,
            sced_instance=sced_instance,
            hour=current_hour,
        )

        # actual tracking
        implemented_profiles = self.tracker.track_market_dispatch(
            market_dispatch=market_signals, date=current_date, hour=current_hour
        )

        # update models
        self.tracker.update_model(**implemented_profiles)
        self.bidder.update_real_time_model(**implemented_profiles)

        return

    def update_observed_dispatch(self, options, simulator, ops_stats):
        """
        This methods extract the actual power delivered by the tracking model and
        inform Prescient, so Prescient can use this data to calculate the settlement
        and etc.

        Arguments:
            options: Prescient options from prescient.simulator.config.

            simulator: Prescient simulator.

            ops_stats: Prescient operation statitstic object

        Returns:
            None

        """
        g = self.bidder.bidding_model_object.model_data.gen_name

        # store the dictionaries that the observed/delivered power levels
        observed_dispatch_level_dicts = [
            ops_stats.observed_thermal_dispatch_levels,
            ops_stats.observed_renewables_levels,
            ops_stats.observed_virtual_dispatch_levels,
        ]

        for observed_dispatch_level in observed_dispatch_level_dicts:
            if g in observed_dispatch_level:
                observed_dispatch_level[g] = self.tracker.get_last_delivered_power()

        return

    def activate_pending_DA_data(self, options, simulator):
        """
        This function puts the day-ahead data computed in the day before into effect,
        i.e. the data for the next day become the data for the current day.

        Arguments:
            options: Prescient options from prescient.simulator.config.

            simulator: Prescient simulator.

        Returns:
            None
        """

        self.current_bids, self.next_bids = self.next_bids, None
        self.current_DA_prices, self.next_DA_prices = self.next_DA_prices, None
        self.current_DA_dispatches, self.next_DA_dispatches = (
            self.next_DA_dispatches,
            None,
        )

        # update the available DA signals
        self.current_avail_DA_prices = self.current_DA_prices
        self.current_avail_DA_dispatches = self.current_DA_dispatches

    def write_plugin_results(self, options, simulator):
        """
        After the simulation is completed, the plugins can write their own customized
        results. Each plugin will have to have a method named 'write_results'.

        Arguments:
            options: Prescient options from prescient.simulator.config.

            simulator: Prescient simulator.

        Returns:
            None

        """

        self.bidder.write_results(path=options.output_directory)
        self.tracker.write_results(path=options.output_directory)

        return



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Author: Andrew Lee
"""
import gc
import pytest

from pyomo.util.check_units import assert_units_consistent
from pyomo.common.timing import TicTocTimer
from pyomo.environ import assert_optimal_termination
import pyomo.common.unittest as unittest

from idaes.core.solvers import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom

# Get default solver for testing
solver = get_solver()


class PerformanceBaseClass:
    """
    Base class for running the IDAES Performance Test Suite.

    This class is intended to be used as a basis for constructing tests to be run as
    part of the IDAES performance testing suite, and should implement the following
    methods:

    * build_model - should return a complete instance of the model being tested.
    * test_performance - a test method for profiling the model of interest which logs
        some performance data. A standard method is provided which logs some common
        performance data, but developers should overload this as required.
    * initialize_model - (optional) a method to initialize the model of interest. This
        method is called as part of the standard test_performance method.
    * solve_model - (optional) a method to solve the model of interest. This method is
        called as part of the standard test_performance method, and a standard method
        is provided by the base class.

    Developers may add additional test methods to the derived class to run additional
    performance tests if desired. These methods should use standard pytest marks and
    naming conventions.

    Usage:

    Performance tests should inherit from this based class and unittest.TestCase.
    Inheritance from TestCase is used to make the derived classes easily discoverable
    for use outside the performance testing infrastructure. The derived class should
    be decorated with the 'performance' pytest mark. This will automatically categorize
    all tests in this class as performance tests that will only be run when explicitly
    called for.

    Raises:
        TypeError is derived class does not inherit from unittest.TestCase.
    """

    # Add flag for skipping unit consistency tests
    # This is required for DAE models for the time being
    # TODO: Fix this once the Pyomo bug is resolved
    TEST_UNITS = True

    def __init_subclass__(cls):
        if not issubclass(cls, unittest.TestCase):
            raise TypeError(
                "Classes derived from PerformanceBaseClass must also inherit from "
                "unittest.TestCase."
            )

    def recordData(self, name, value):
        """
        A method for recording data associated with a test.
        """
        tmp = getattr(self, "testdata", None)
        if tmp is not None:
            tmp[name] = value

    def build_model(self):
        """
        Return a fully constructed version of the model of interest. Developers must
        overload this method.

        Args:
            None

        Returns:
            None
        """
        raise NotImplementedError(
            "Test class has not implemented a build_model method."
        )

    def initialize_model(self, model):
        """
        Initialize provided model. Developers should overload this method to provide
        model specific initialization instructions.

        Args:
            model - constructed model object to be initialized

        Returns:
            None
        """
        raise NotImplementedError(
            "Test class has not implemented an initialize_model method."
        )

    def solve_model(self, model):
        """
        Solve provided model. Developers should overload this if necessary to customize
        solver and arguments.

        Args:
            model - constructed model object to be solved

        Returns:
            None
        """
        results = solver.solve(model)

        assert_optimal_termination(results)

    @pytest.mark.performance
    def test_performance(self):
        """
        Standard method for executing performance test runs.

        This method calls the build_model, initialize_model and solve_model methods
        thus developers must provide these. This method logs the following information:

        1. Model construction time
        2. Time required to assert unit consistency
        3. Model initialization time
        4. Model solution time (post-initialization)

        Developers may overload this method to customize the information recorded.
        Note that dynamic models currently fail unit consistency checks, and thus
        this method will fail if used on a dynamic model (for now).

        Args:
            None
        """
        # Build model and record execution time
        gc.collect()
        timer = TicTocTimer()
        model = self.build_model()
        self.recordData("build model", timer.toc("build model"))
        assert degrees_of_freedom(model) == 0

        # Check unit consistency and record execution time
        if self.TEST_UNITS:
            gc.collect()
            timer.tic(None)
            assert_units_consistent(model)
            self.recordData("unit consistency", timer.toc("unit consistency"))

        # Initialize model and record execution time
        gc.collect()
        timer.tic(None)
        self.initialize_model(model)
        self.recordData("initialize", timer.toc("initialize"))

        # Solve model and record execution time
        gc.collect()
        timer.tic(None)
        self.solve_model(model)
        self.recordData("final solve", timer.toc("final solve"))



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Costing package based on methods from:

    Process and Product Design Principles: Synthesis, Analysis, and
    Evaluation
    Seider, Seader, Lewin, Windagdo, 3rd Ed. John Wiley and Sons
    Chapter 22. Cost Accounting and Capital Cost Estimation
    22.2 Cost Indexes and Capital Investment

Currently, this costing package only includes methods for capital costing of
unit operations.
"""
# TODO: Missing docstrings
# pylint: disable=missing-class-docstring

import pyomo.environ as pyo

# TODO: HX1D not supported - does not define area (has shell_area & tube_area)
from idaes.models.unit_models import (
    Compressor,
    CSTR,
    Flash,
    Heater,
    HeatExchanger,
    HeatExchangerNTU,
    PFR,
    PressureChanger,
    Pump,
    StoichiometricReactor,
    Turbine,
)
from idaes.models.unit_models.pressure_changer import ThermodynamicAssumption
from idaes.models.unit_models.heat_exchanger import HeatExchangerFlowPattern

from idaes.core import declare_process_block_class
from idaes.core.util.exceptions import ConfigurationError
from idaes.core.util.constants import Constants
from idaes.core.util.math import smooth_max
from idaes.core.util.misc import StrEnum

from idaes.core import (
    FlowsheetCostingBlockData,
    register_idaes_currency_units,
)

import idaes.logger as idaeslog

_log = idaeslog.getLogger(__name__)

# Some more information about this module
__author__ = "Miguel Zamarripa, Andrew Lee"


class HXType(StrEnum):
    floating_head = "floating_head"
    fixed_head = "fixed_head"
    Utube = "Utube"
    kettle_vap = "Kettle_vap"


class HXMaterial(StrEnum):
    CarbonSteelCarbonSteel = "CarbonSteelCarbonSteel"
    CarbonSteelBrass = "CarbonSteelBrass"
    CarbonSteelStainlessSteel = "CarbonSteelStainlessSteel"
    CarbonSteelMonel = "CarbonSteelMonel"
    CarbonSteelTitanium = "CarbonSteelTitanium"
    CarbonSteelCrMoSteel = "CarbonSteelCrMoSteel"
    CrMoSteelCrMoSteel = "CrMoSteelCrMoSteel"
    StainlessSteelStainlessSteel = "StainlessSteelStainlessSteel"
    MonelMonel = "MonelMonel"
    TitaniumTitanium = "TitaniumTitanium"


class HXTubeLength(StrEnum):
    EightFoot = "8ft"
    TwelveFoot = "12ft"
    SixteenFoot = "16ft"
    TwentyFoot = "20ft"


class VesselMaterial(StrEnum):
    CarbonSteel = "Carbon_steel"
    LowAlloySteel = "LowAlloySteel"
    StainlessSteel304 = "StainlessSteel304"
    StainlessSteel316 = "StainlessSteel316"
    Carpenter20CB3 = "Carpenter20CB3"
    Nickel200 = "Nickel200"
    Monel400 = "Monel400"
    Inconel600 = "Inconel600"
    Incoloy825 = "Incoloy825"
    Titanium = "Titanium"


class TrayType(StrEnum):
    Sieve = "Sieve"
    Valve = "Valve"
    BubbleCap = "BubbleCap"


class TrayMaterial(StrEnum):
    CarbonSteel = "CarbonSteel"
    StainlessSteel303 = "StainlessSteel303"
    StainlessSteel316 = "StainlessSteel316"
    Carpenter20CB3 = "Carpenter20CB3"
    Monel = "Monel"


class HeaterMaterial(StrEnum):
    CarbonSteel = "CarbonSteel"
    CrMoSteel = "CrMoSteel"
    StainlessSteel = "StainlessSteel"


class HeaterSource(StrEnum):
    Fuel = "Fuel"
    Reformer = "Reformer"
    Pyrolysis = "Pyrolysis"
    HotWater = "HotWater"
    Salts = "Salts"
    DowthermA = "DowthermA"
    steamBoiler = "SteamBoiler"


class CompressorType(StrEnum):
    Centrifugal = "Centrifugal"
    Reciprocating = "Reciprocating"
    Screw = "Screw"


class CompressorDriveType(StrEnum):
    ElectricMotor = "ElectricMotor"
    SteamTurbine = "SteamTurbine"
    gasTurbine = "GasTurbine"


class CompressorMaterial(StrEnum):
    CarbonSteel = "CarbonSteel"
    StainlessSteel = "StainlessSteel"
    NickelAlloy = "NickelAlloy"


class PumpMaterial(StrEnum):
    CastIron = "CastIron"
    DuctileIron = "DuctileIron"
    CastSteel = "CastSteel"
    Bronze = "Bronze"
    StainlessSteel = "StainlessSteel"
    HastelloyC = "HastelloyC"
    Monel = "Monel"
    Nickel = "Nickel"
    Titanium = "Titanium"
    NiAlBronze = "NiAlBronze"
    CarbonSteel = "CarbonSteel"


class PumpType(StrEnum):
    Centrifugal = "Centrifugal"
    ExternalGear = "ExternalGear"
    Reciprocating = "Reciprocating"


class PumpMotorType(StrEnum):
    Open = "open"
    Enclosed = "enclosed"
    ExplosionProof = "explosion_proof"


class FanType(StrEnum):
    CentrifugalBackward = "CentrifugalBackward"
    CentrifugalStraight = "CentrifugalStraight"
    VaneAxial = "VaneAxial"
    TubeAxial = "TubeAxial"


class FanMaterial(StrEnum):
    CarbonSteel = "CarbonSteel"
    Fiberglass = "Fiberglass"
    StainlessSteel = "StainlessSteel"
    NickelAlloy = "NickelAlloy"


class BlowerType(StrEnum):
    Centrifugal = "Centrifugal"
    Rotary = "Rotary"


class BlowerMaterial(StrEnum):
    CarbonSteel = "CarbonSteel"
    Aluminum = "Aluminum"
    Fiberglass = "Fiberglass"
    StainlessSteel = "StainlessSteel"
    NickelAlloy = "NickelAlloy"


@declare_process_block_class("SSLWCosting")
class SSLWCostingData(FlowsheetCostingBlockData):
    # Register currency and conversion rates based on CE Index
    register_idaes_currency_units()

    def build_global_params(self):
        """
        This is where we can declare any global parameters we need, such as
        Lang factors, or coefficients for costing methods that should be
        shared across the process.

        You can do what you want here, so you could have e.g. sub-Blocks
        for each costing method to separate the parameters for each method.
        """
        # Set the base year for all costs
        self.base_currency = pyo.units.USD_2018
        # Set a base period for all operating costs
        self.base_period = pyo.units.year

    def build_process_costs(self):
        """
        This is where you do all your process wide costing.
        This is completely up to you, but you will have access to the
        following aggregate costs:

        1. self.aggregate_capital_cost
        2. self.aggregate_fixed_operating_cost
        3. self.aggregate_variable_operating_cost
        4. self.aggregate_flow_costs (indexed by flow type)
        """
        # TODO: Do we have any process level methods to add here?

    @staticmethod
    def initialize_build(*args, **kwargs):
        """
        Here we can add initialization steps for the things we built in
        build_process_costs.

        Note that the aggregate costs will be initialized by the framework.
        """
        # TODO: For now, no additional process level costs to initialize

    def cost_heat_exchanger(
        blk,
        hx_type=HXType.Utube,
        material_type=HXMaterial.StainlessSteelStainlessSteel,
        tube_length=HXTubeLength.TwelveFoot,
        integer=True,
    ):
        """
        Heat exchanger costing method.

        This method computes the purchase cost (CP) for a shell and tube heat
        exchanger (Eq. 22.43), the model computes the base cost (CB for 4 types
        of heat exchangers, such as floating head, fixed head, U-tube, and
        Kettle vaporizer), construction material factor (mat_factor),
        pressure design factor (pressure_factor), and
        tube length correction factor (length_factor),
        using Chemical Engineering base cost index of 500.

        Purchase Cost = pressure_factor *
                        material_factor *
                        length_factor *
                        base_cost_per_unit *
                        number_of_units

        Args:
            hx_type: HXType Enum indicating type of heat exchanger design,
                default = HXType.Utube.
            material_type: HXMaterial Enum indicating material of construction,
                default = HXMaterial.StainlessSteelStainlessSteel.
            tube_length: HXTubeLength Enum indicating length of HX tubes,
                default = HXTubeLength.TwelveFoot.
            integer: whether the number of units should be constrained to be
                an integer or not (default = True).
        """
        # Validate arguments
        if hx_type not in HXType:
            raise ConfigurationError(
                f"{blk.unit_model.name} received invalid argument for hx_type:"
                f" {hx_type}. Argument must be a member of the HXType Enum."
            )
        if material_type not in HXMaterial:
            raise ConfigurationError(
                f"{blk.unit_model.name} received invalid argument for "
                f"material_type: {material_type}. Argument must be a member "
                "of the HXMaterial Enum."
            )
        if tube_length not in HXTubeLength:
            raise ConfigurationError(
                f"{blk.unit_model.name} received invalid argument for "
                f"tube_length: {tube_length}. Argument must be a member "
                "of the HXTubeLength Enum."
            )

        # Build generic costing variables
        _make_common_vars(blk, integer)

        # Length correction factor
        c_fl = {"8ft": 1.25, "12ft": 1.12, "16ft": 1.05, "20ft": 1.00}
        blk.length_factor = pyo.Param(
            mutable=True,
            initialize=c_fl[tube_length],
            doc="HX tube length correction factor",
        )

        blk.hx_oversize = pyo.Param(
            mutable=True,
            initialize=1.1,
            doc="HX oversize factor (1.1 to 1.5)",
            units=pyo.units.ft**-2,
        )

        # --------------------------------------------------
        # Base cost calculation - based on selected heat exchanger type:
        alpha_dict = {
            HXType.floating_head: {1: 11.9052, 2: 0.8709, 3: 0.09005},
            HXType.fixed_head: {1: 11.2927, 2: 0.8228, 3: 0.09861},
            HXType.Utube: {1: 11.3852, 2: 0.9186, 3: 0.09790},
            HXType.kettle_vap: {1: 12.2052, 2: 0.8709, 3: 0.09005},
        }
        alpha = alpha_dict[hx_type]

        # Convert area to square feet
        area_unit = (
            pyo.units.convert(blk.unit_model.area, to_units=pyo.units.ft**2)
            / blk.number_of_units
        )

        @blk.Constraint()
        def base_cost_per_unit_eq(blk):
            return (
                blk.base_cost_per_unit
                == pyo.exp(
                    alpha[1]
                    - alpha[2] * pyo.log(area_unit * blk.hx_oversize)
                    + alpha[3] * pyo.log(area_unit * blk.hx_oversize) ** 2
                )
                * pyo.units.USD_CE500
            )

        @blk.Expression(doc="Base cost for all installed units")
        def base_cost(blk):
            return blk.base_cost_per_unit * blk.number_of_units

        # ------------------------------------------------------
        # Material of construction factor Eq. 22.44 in the reference
        blk.material_factor = pyo.Var(
            initialize=3.5,
            domain=pyo.NonNegativeReals,
            doc="Construction material correction factor",
        )

        hx_material_factor_dict = {
            HXMaterial.CarbonSteelCarbonSteel: {"A": 0.00, "B": 0.00},
            HXMaterial.CarbonSteelBrass: {"A": 1.08, "B": 0.05},
            HXMaterial.CarbonSteelStainlessSteel: {"A": 1.75, "B": 0.13},
            HXMaterial.CarbonSteelMonel: {"A": 2.10, "B": 0.13},
            HXMaterial.CarbonSteelTitanium: {"A": 5.20, "B": 0.16},
            HXMaterial.CarbonSteelCrMoSteel: {"A": 1.55, "B": 0.05},
            HXMaterial.CrMoSteelCrMoSteel: {"A": 1.70, "B": 0.07},
            HXMaterial.StainlessSteelStainlessSteel: {"A": 2.70, "B": 0.07},
            HXMaterial.MonelMonel: {"A": 3.30, "B": 0.08},
            HXMaterial.TitaniumTitanium: {"A": 9.60, "B": 0.06},
        }

        mf = hx_material_factor_dict[material_type]

        @blk.Constraint()
        def hx_material_eqn(self):
            if material_type == HXMaterial.CarbonSteelCarbonSteel:
                return blk.material_factor == 1
            else:
                return (
                    blk.material_factor
                    == mf["A"] + (area_unit / (100 * pyo.units.ft**2)) ** mf["B"]
                )

        # ------------------------------------------------------
        # Pressure factor calculation
        t0 = blk.unit_model.flowsheet().time.first()

        # Assume higher pressure fluid is tube side
        blk.pressure_factor = pyo.Var(
            initialize=1, domain=pyo.NonNegativeReals, doc="Pressure design factor"
        )

        try:
            tube_props = blk.unit_model.hot_side.properties_in[t0]
        except AttributeError:
            # Assume HX1D
            if blk.unit_model.config.flow_type == HeatExchangerFlowPattern.cocurrent:
                inlet_x = 0
            else:
                inlet_x = 1
            tube_props = blk.unit_model.tube.properties[0, inlet_x]

        # Pressure units must be in psig
        pressure = pyo.units.convert(
            tube_props.pressure, to_units=pyo.units.psi
        ) - pyo.units.convert(1 * pyo.units.atm, to_units=pyo.units.psi)

        @blk.Constraint()
        def p_factor_eq(blk):
            # Equation valid from 600 pisg to 3000 psig
            # return self.pressure_factor == (
            #     0.8510 + 0.1292*(pressure/600) + 0.0198*(pressure/600)**2)
            # Equation valid from 100 pisg to 2000 psig
            return blk.pressure_factor == (
                0.9803
                + 0.0180 * (pressure / (100 * pyo.units.psi))
                + 0.0017 * (pressure / (100 * pyo.units.psi)) ** 2
            )

        # Total capital cost equation
        @blk.Constraint()
        def capital_cost_constraint(blk):
            return blk.capital_cost == (
                blk.pressure_factor
                * blk.material_factor
                * blk.length_factor
                * blk.base_cost
            )

    def cost_vessel(
        blk,
        vertical=False,
        material_type=VesselMaterial.CarbonSteel,
        shell_thickness=1.25 * pyo.units.inch,
        weight_limit=1,
        aspect_ratio_range=1,
        include_platforms_ladders=True,
        vessel_diameter=None,
        vessel_length=None,
        number_of_units=1,
        number_of_trays=None,
        tray_material=TrayMaterial.CarbonSteel,
        tray_type=TrayType.Sieve,
    ):
        """
        Generic vessel costing method.

        Args:
            vertical: alignment of vessel; vertical if True, horizontal if
                False (default=False).
            material_type: VesselMaterial Enum indicating material of
                construction, default = VesselMaterial.CarbonSteel.
            shell_thickness: thickness of vessel shell, including pressure
                allowance. Default = 1.25 inches.
            weight_limit: 1: (default) 1000 to 920,000 lb, 2: 4200 to 1M lb.
                Option 2 is only valid for vertical vessels.
            aspect_ratio_range: vertical vessels only, default = 1;
                1: 3 < D < 21 ft, 12 < L < 40 ft, 2: 3 < D < 24 ft; 27 < L < 170 ft.
            include_platforms_ladders: whether to include platforms and
                ladders in costing , default = True.
            vessel_diameter: Pyomo component representing vessel diameter.
                If not provided, assumed to be named "diameter"
            vessel_length: Pyomo component representing vessel length.
                If not provided, assumed to be named "length".
            number_of_units: Integer or Pyomo component representing the
                number of parallel units to be costed, default = 1.
            number_of_trays: Pyomo component representing the number of
                distillation trays in vessel (default=None)
            tray_material: Only required if number_of_trays is not None.
                TrayMaterial Enum indicating material of construction for
                distillation trays, default = TrayMaterial.CarbonSteel.
            tray_type: Only required if number_of_trays is not None.
                TrayType Enum indicating type of distillation trays to use,
                default = TrayMaterial.Sieve.
        """
        # Build generic costing variables
        blk.base_cost_per_unit = pyo.Var(
            initialize=1e5,
            domain=pyo.NonNegativeReals,
            units=pyo.units.USD_CE500,
            doc="Base cost per unit",
        )

        blk.capital_cost = pyo.Var(
            initialize=1e4,
            domain=pyo.NonNegativeReals,
            bounds=(0, None),
            units=pyo.units.USD_CE500,
            doc="Capital cost of all units",
        )

        # Check arguments
        if material_type not in VesselMaterial:
            raise ConfigurationError(
                f"{blk.unit_model.name} received invalid argument for "
                f"material_type: {material_type}. Argument must be a member "
                "of the VesselMaterial Enum."
            )

        if not vertical and number_of_trays is not None:
            raise ConfigurationError(
                f"{blk.unit_model.name} distillation trays are only supported "
                "for vertical vessels."
            )

        if weight_limit == 2 and not vertical:
            raise ConfigurationError(
                f"{blk.unit_model.name} weight_limit option 2 is only valid "
                "for vertical vessels."
            )
        elif weight_limit not in [1, 2]:
            raise ConfigurationError(
                f"{blk.unit_model.name} weight_limit argument must be 1 or 2;"
                f" received {weight_limit}."
            )

        # Check references to diameter and length
        if vessel_diameter is None:
            try:
                vessel_diameter = blk.unit_model.diameter
            except AttributeError:
                raise ConfigurationError(
                    f"{blk.unit_model.name} does not have a component named "
                    "diameter. Please provide a reference to the vessel "
                    "diameter as an argument to cost_unit."
                )

        if vessel_length is None:
            try:
                vessel_length = blk.unit_model.length
            except AttributeError:
                raise ConfigurationError(
                    f"{blk.unit_model.name} does not have a component named "
                    "length. Please provide a reference to the vessel "
                    "length as an argument to cost_unit."
                )

        D_in = pyo.units.convert(vessel_diameter, to_units=pyo.units.inch)
        L_in = pyo.units.convert(vessel_length, to_units=pyo.units.inch)

        # Material densities in lb/cubic inch
        material_factor_dict = {
            VesselMaterial.CarbonSteel: {"factor": 1.0, "density": 0.284},
            VesselMaterial.LowAlloySteel: {"factor": 1.2, "density": 0.271},
            VesselMaterial.StainlessSteel304: {"factor": 1.7, "density": 0.270},
            VesselMaterial.StainlessSteel316: {"factor": 2.1, "density": 0.276},
            VesselMaterial.Carpenter20CB3: {"factor": 3.2, "density": 0.29},
            VesselMaterial.Nickel200: {"factor": 5.4, "density": 0.3216},
            VesselMaterial.Monel400: {"factor": 3.6, "density": 0.319},
            VesselMaterial.Inconel600: {"factor": 3.9, "density": 0.3071},
            VesselMaterial.Incoloy825: {"factor": 3.7, "density": 0.2903},
            VesselMaterial.Titanium: {"factor": 7.7, "density": 0.1628},
        }
        material_factors = material_factor_dict[material_type]

        # Users should calculate the pressure design based shell thickness
        # Pressure factor assumed to be included in thickness
        blk.shell_thickness = pyo.Param(
            mutable=True, doc="Shell thickness", units=pyo.units.inch
        )
        blk.shell_thickness.set_value(shell_thickness)
        blk.material_factor = pyo.Param(
            initialize=material_factors["factor"],
            mutable=True,
            doc="Construction material correction factor",
        )
        blk.material_density = pyo.Param(
            initialize=material_factors["density"],
            mutable=True,
            doc="Density of the metal",
            units=pyo.units.pound / pyo.units.inch**3,
        )

        # Calculate weight of vessel
        blk.weight = pyo.Var(
            initialize=1000,
            domain=pyo.NonNegativeReals,
            doc="Weight of vessel in lb",
            units=pyo.units.pound,
        )

        @blk.Constraint()
        def weight_eq(blk):
            return blk.weight == (
                Constants.pi
                * (D_in + blk.shell_thickness)
                * (L_in + 0.8 * D_in)
                * blk.shell_thickness
                * blk.material_density
            )

        # Base Vessel cost
        # Alpha factors for correlation
        # 2nd key is weight_limit option
        alpha_dict = {
            "H": {1: {1: 8.9552, 2: -0.2330, 3: 0.04333}, 2: None},
            "V": {
                1: {1: 7.0132, 2: 0.18255, 3: 0.02297},
                2: {1: 7.2756, 2: 0.18255, 3: 0.02297},
            },
        }

        if vertical:
            alpha = alpha_dict["V"][weight_limit]
        else:
            alpha = alpha_dict["H"][weight_limit]

        @blk.Constraint()
        def base_cost_constraint(blk):
            return blk.base_cost_per_unit == (
                pyo.exp(
                    alpha[1]
                    + alpha[2] * (pyo.log(blk.weight / pyo.units.pound))
                    + alpha[3] * (pyo.log(blk.weight / pyo.units.pound) ** 2)
                )
                * pyo.units.USD_CE500
            )

        # Add platform and ladder costs if required
        if include_platforms_ladders:
            SSLWCostingData._cost_platforms_ladders(
                blk,
                vertical=vertical,
                aspect_ratio_range=aspect_ratio_range,
                vessel_diameter=vessel_diameter,
                vessel_length=vessel_length,
            )

        # Add distillation trays costs if required
        if number_of_trays is not None:
            SSLWCostingData._cost_distillation_trays(
                blk,
                tray_material=tray_material,
                tray_type=tray_type,
                vessel_diameter=vessel_diameter,
                number_of_trays=number_of_trays,
            )

        # Total capital cost of vessel and ancillary equipment
        @blk.Constraint()
        def capital_cost_constraint(blk):
            cost_expr = blk.material_factor * blk.base_cost_per_unit
            if include_platforms_ladders:
                cost_expr += blk.base_cost_platforms_ladders
            if number_of_trays is not None:
                cost_expr += blk.base_cost_trays

            return blk.capital_cost == cost_expr * number_of_units

    def _cost_platforms_ladders(
        blk, vertical, aspect_ratio_range, vessel_diameter, vessel_length
    ):
        """
        Method for calculating costs of platforms and ladders
        """
        blk.base_cost_platforms_ladders = pyo.Var(
            initialize=1000,
            units=pyo.units.USD_CE500,
            domain=pyo.NonNegativeReals,
            doc="Base cost of platforms and ladders",
        )

        D = pyo.units.convert(vessel_diameter, to_units=pyo.units.foot)
        L = pyo.units.convert(vessel_length, to_units=pyo.units.foot)

        if not vertical:

            @blk.Constraint()
            def cost_platforms_ladders_eq(blk):
                return blk.base_cost_platforms_ladders == (
                    2005 * (D / pyo.units.foot) ** 0.20294 * pyo.units.USD_CE500
                )

        else:

            @blk.Constraint()
            def cost_platforms_ladders_eq(blk):
                if aspect_ratio_range == 1:
                    return blk.base_cost_platforms_ladders == (
                        361.8
                        * (D / pyo.units.foot) ** 0.73960
                        * (L / pyo.units.foot) ** 0.70684
                        * pyo.units.USD_CE500
                    )
                elif aspect_ratio_range == 2:
                    return blk.base_cost_platforms_ladders == (
                        309.9
                        * (D / pyo.units.foot) ** 0.63316
                        * (L / pyo.units.foot) ** 0.80161
                        * pyo.units.USD_CE500
                    )
                else:
                    raise ConfigurationError(
                        f"{blk.unit_model.name} received invalid value for "
                        f"aspect_ratio_range argument: {aspect_ratio_range}. "
                        "Value must be 1 or 2."
                    )

    def _cost_distillation_trays(
        blk, tray_material, tray_type, vessel_diameter, number_of_trays
    ):
        # Check arguments
        if tray_material not in TrayMaterial:
            raise ConfigurationError(
                f"{blk.unit_model.name} received invalid argument for "
                f"tray_material: {tray_material}. Argument must be a member "
                "of the TrayMaterial Enum."
            )
        if tray_type not in TrayType:
            raise ConfigurationError(
                f"{blk.unit_model.name} received invalid argument for "
                f"tray_type: {tray_type}. Argument must be a member "
                "of the TrayType Enum."
            )

        D = pyo.units.convert(vessel_diameter, to_units=pyo.units.foot)

        blk.base_cost_trays = pyo.Var(
            initialize=1e6,
            units=pyo.units.USD_CE500,
            domain=pyo.NonNegativeReals,
            doc="Purchase cost of trays",
        )

        tray_type_dict = {
            TrayType.Sieve: 1,
            TrayType.Valve: 1.18,
            TrayType.BubbleCap: 1.87,
        }
        blk.tray_type_factor = pyo.Param(
            initialize=tray_type_dict[tray_type], mutable=True, doc="Tray type factor"
        )

        blk.tray_material_factor = pyo.Var(
            initialize=1.0, doc="FTM material of construction factor for trays"
        )
        # Alpha parameters for material factor
        alpha = {
            TrayMaterial.CarbonSteel: {1: 1, 2: 0},
            TrayMaterial.StainlessSteel303: {1: 1.189, 2: 0.0577},
            TrayMaterial.StainlessSteel316: {1: 1.401, 2: 0.0724},
            TrayMaterial.Carpenter20CB3: {1: 1.525, 2: 0.0788},
            TrayMaterial.Monel: {1: 2.306, 2: 0.1120},
        }

        # Calculating tray factor value
        # Column diameter in ft, eqn. valid for 2 to 16 ft
        @blk.Constraint()
        def tray_material_factor_eq(blk):
            return blk.tray_material_factor == (
                alpha[tray_material][1] + alpha[tray_material][2] * D / pyo.units.foot
            )

        # Calculate cost factor for number of trays
        blk.number_trays_factor = pyo.Var(
            initialize=1,
            units=pyo.units.dimensionless,
            doc="Cost factor for number of trays",
        )

        @blk.Constraint()
        def num_tray_factor_constraint(blk):
            return blk.number_trays_factor == smooth_max(
                1, 2.25 / (1.0414**number_of_trays)
            )

        # Calculate base cost of a single tray
        blk.base_cost_per_tray = pyo.Var(
            initialize=1e4,
            domain=pyo.NonNegativeReals,
            units=pyo.units.USD_CE500,
            doc="Base cost of a single tray",
        )

        @blk.Constraint()
        def single_tray_cost_constraint(blk):
            return blk.base_cost_per_tray == (
                468.00 * pyo.exp(0.1739 * D / pyo.units.foot) * pyo.units.USD_CE500
            )

        # Capital cost of trays
        @blk.Constraint()
        def tray_costing_constraint(blk):
            return blk.base_cost_trays == (
                number_of_trays
                * blk.number_trays_factor
                * blk.tray_type_factor
                * blk.tray_material_factor
                * blk.base_cost_per_tray
            )

    def cost_vertical_vessel(
        blk,
        material_type=VesselMaterial.CarbonSteel,
        shell_thickness=1.25 * pyo.units.inch,
        weight_limit=1,
        aspect_ratio_range=1,
        include_platforms_ladders=True,
        vessel_diameter=None,
        vessel_length=None,
        number_of_units=1,
        number_of_trays=None,
        tray_material=TrayMaterial.CarbonSteel,
        tray_type=TrayType.Sieve,
    ):
        """
        Specific case of vessel costing method for vertical vessels.

        Args:
            material_type: VesselMaterial Enum indicating material of
                construction, default = VesselMaterial.CarbonSteel.
            shell_thickness: thickness of vessel shell, including pressure
                allowance. Default = 1.25 inches.
            weight_limit: 1: (default) 1000 to 920,000 lb, 2: 4200 to 1M lb.
            aspect_ratio_range: default = 1;
                1: 3 < D < 21 ft, 12 < L < 40 ft, 2: 3 < D < 24 ft; 27 < L < 170 ft.
            include_platforms_ladders: whether to include platforms and
                ladders in costing , default = True.
            vessel_diameter: Pyomo component representing vessel diameter.
                If not provided, assumed to be named "diameter"
            vessel_length: Pyomo component representing vessel length.
                If not provided, assumed to be named "length".
            number_of_units: Integer or Pyomo component representing the
                number of parallel units to be costed, default = 1.
            number_of_trays: Pyomo component representing the number of
                distillation trays in vessel (default=None)
            tray_material: Only required if number_of_trays is not None.
                TrayMaterial Enum indicating material of construction for
                distillation trays, default = TrayMaterial.CarbonSteel.
            tray_type: Only required if number_of_trays is not None.
                TrayType Enum indicating type of distillation trays to use,
                default = TrayMaterial.Sieve.
        """
        SSLWCostingData.cost_vessel(
            blk,
            vertical=True,
            material_type=VesselMaterial.CarbonSteel,
            shell_thickness=1.25 * pyo.units.inch,
            weight_limit=1,
            aspect_ratio_range=1,
            include_platforms_ladders=True,
            vessel_diameter=None,
            vessel_length=None,
            number_of_units=1,
            number_of_trays=None,
            tray_material=TrayMaterial.CarbonSteel,
            tray_type=TrayType.Sieve,
        )

    def cost_horizontal_vessel(
        blk,
        material_type=VesselMaterial.CarbonSteel,
        shell_thickness=1.25 * pyo.units.inch,
        include_platforms_ladders=True,
        vessel_diameter=None,
        vessel_length=None,
        number_of_units=1,
    ):
        """
        Specific case of vessel costing method for horizontal vessels.

        Arguments which do not apply to horizontal vessels are excluded.

        Args:
            material_type: VesselMaterial Enum indicating material of
                construction, default = VesselMaterial.CarbonSteel.
            shell_thickness: thickness of vessel shell, including pressure
                allowance. Default = 1.25 inches.
            include_platforms_ladders: whether to include platforms and
                ladders in costing, default = True.
            vessel_diameter: Pyomo component representing vessel diameter.
                If not provided, assumed to be named "diameter"
            vessel_length: Pyomo component representing vessel length.
                If not provided, assumed to be named "length".
            number_of_units: Integer or Pyomo component representing the
                number of parallel units to be costed, default = 1.
        """
        SSLWCostingData.cost_vessel(
            blk,
            vertical=False,
            material_type=VesselMaterial.CarbonSteel,
            shell_thickness=1.25 * pyo.units.inch,
            weight_limit=1,
            include_platforms_ladders=True,
            vessel_diameter=None,
            vessel_length=None,
            number_of_units=1,
        )

    def cost_fired_heater(
        blk,
        heat_source=HeaterSource.Fuel,
        material_type=HeaterMaterial.CarbonSteel,
        integer=True,
    ):
        """
        Generic costing method for fired heaters.

        Args:
            heat_source: HeaterSource Enum indicating type of source of heat,
                default = HeaterSource.Fuel.
            material_type: HeaterMaterial Enum indicating material of
                construction, default = HeaterMaterial.CarbonSteel.
            integer: whether the number of units should be constrained to be
                an integer or not (default = True).
        """
        # Validate arguments
        if heat_source not in HeaterSource:
            raise ConfigurationError(
                f"{blk.unit_model.name} received invalid argument for "
                f"heat_source: {heat_source}. Argument must be a member of "
                "the HeaterSource Enum."
            )
        if material_type not in HeaterMaterial:
            raise ConfigurationError(
                f"{blk.unit_model.name} received invalid argument for "
                f"material_type: {material_type}. Argument must be a member "
                "of the HeaterMaterial Enum."
            )

        # Build generic costing variables
        _make_common_vars(blk, integer)

        # Convert pressure to psi,g
        t0 = blk.unit_model.flowsheet().time.first()

        P = pyo.units.convert(
            blk.unit_model.control_volume.properties_in[t0].pressure,
            to_units=pyo.units.psi,
        ) - pyo.units.convert(1 * pyo.units.atm, to_units=pyo.units.psi)

        # Convert heat duty to BTU/hr
        Q = (
            pyo.units.convert(
                blk.unit_model.heat_duty[t0], to_units=pyo.units.BTU / pyo.units.hr
            )
            / blk.number_of_units
        )

        # Material factor
        material_factor_dict = {
            HeaterMaterial.CarbonSteel: 1.0,
            HeaterMaterial.CrMoSteel: 1.4,
            HeaterMaterial.StainlessSteel: 1.7,
        }
        blk.material_factor = pyo.Param(
            initialize=material_factor_dict[material_type],
            domain=pyo.NonNegativeReals,
            doc="Construction material correction factor",
        )

        # Pressure design factor calculation
        blk.pressure_factor = pyo.Var(
            initialize=1.1, domain=pyo.NonNegativeReals, doc="Pressure design factor"
        )

        @blk.Constraint()
        def pressure_factor_eq(blk):
            return blk.pressure_factor == (
                0.986
                - 0.0035 * (P / (500.00 * pyo.units.psi))
                + 0.0175 * (P / (500.00 * pyo.units.psi)) ** 2
            )

        @blk.Constraint()
        def base_cost_per_unit_eq(blk):
            if heat_source == HeaterSource.Fuel:
                bc_expr = pyo.exp(
                    0.32325 + 0.766 * pyo.log(Q / pyo.units.BTU * pyo.units.hr)
                )
            elif heat_source == HeaterSource.Reformer:
                bc_expr = 0.859 * (Q / pyo.units.BTU * pyo.units.hr) ** 0.81
            elif heat_source == HeaterSource.Pyrolysis:
                bc_expr = 0.650 * (Q / pyo.units.BTU * pyo.units.hr) ** 0.81
            elif heat_source == HeaterSource.HotWater:
                bc_expr = pyo.exp(
                    9.593
                    - 0.3769 * pyo.log((Q / pyo.units.BTU * pyo.units.hr))
                    + 0.03434 * pyo.log((Q / pyo.units.BTU * pyo.units.hr)) ** 2
                )
            elif heat_source == HeaterSource.Salts:
                bc_expr = 12.32 * (Q / pyo.units.BTU * pyo.units.hr) ** 0.64
            elif heat_source == HeaterSource.DowthermA:
                bc_expr = 12.74 * (Q / pyo.units.BTU * pyo.units.hr) ** 0.65
            elif heat_source == HeaterSource.steamBoiler:
                bc_expr = 0.367 * (Q / pyo.units.BTU * pyo.units.hr) ** 0.77

            return blk.base_cost_per_unit == bc_expr * pyo.units.USD_CE500

        @blk.Expression(doc="Base cost for all units installed")
        def base_cost(blk):
            return blk.base_cost_per_unit * blk.number_of_units

        # Total capital cost of heater(s)
        @blk.Constraint()
        def capital_cost_constraint(blk):
            return blk.capital_cost == (
                blk.material_factor * blk.pressure_factor * blk.base_cost
            )

    def cost_compressor(
        blk,
        compressor_type=CompressorType.Centrifugal,
        drive_type=CompressorDriveType.ElectricMotor,
        material_type=CompressorMaterial.StainlessSteel,
        integer=True,
    ):
        """
        Generic costing method for compressors.

        Args:
            compressor_type: CompressorType Enum indicating type of type of
                equipment, default = CompressorType.Centrifugal.
            material_type: CompressorMaterial Enum indicating material of
                construction, default = CompressorMaterial.StainlessSteel.
            drive_type: CompressorDriveType Enum indicating type of type of
                drive to be used, default = CompressorDriveType.ElectricMotor.
            integer: whether the number of units should be constrained to be
                an integer or not (default = True).
        """
        # Confirm that unit is a compressor
        if not blk.unit_model.config.compressor:
            _log.warning(
                "cost_compressor method is only appropriate for "
                "pressure changers with the compressor argument "
                "equal to True."
            )
        # compressor = True, and using non-isentropic assumption
        # pumps should use the cost_pump method
        if (
            blk.unit_model.config.thermodynamic_assumption
            == ThermodynamicAssumption.pump
        ):
            _log.warning(
                f"{blk.unit_model.name} - pressure changers with the pump "
                "assumption should use the cost_pump method."
            )
        # isothermal or adiabatic compressors are too simple to cost
        elif blk.unit_model.config.thermodynamic_assumption in [
            ThermodynamicAssumption.isothermal,
            ThermodynamicAssumption.adiabatic,
        ]:
            _log.warning(
                f"{blk.unit_model.name} - pressure changers without isentropic "
                "assumption are too simple to be costed."
            )

        # Build generic costing variables
        _make_common_vars(blk, integer)

        work = blk.unit_model.work_mechanical[blk.unit_model.flowsheet().time.first()]

        work_hp = pyo.units.convert(work / blk.number_of_units, to_units=pyo.units.hp)

        # Compressor Purchase Cost Correlation
        FD_param = {
            CompressorDriveType.ElectricMotor: 1,
            CompressorDriveType.SteamTurbine: 1.15,
            CompressorDriveType.gasTurbine: 1.25,
        }
        blk.drive_factor = pyo.Param(
            mutable=True, initialize=FD_param[drive_type], doc="Mover drive factor"
        )

        material_factor_dict = {
            CompressorMaterial.CarbonSteel: 1,
            CompressorMaterial.StainlessSteel: 2.5,
            CompressorMaterial.NickelAlloy: 5.0,
        }
        blk.material_factor = pyo.Param(
            mutable=True,
            initialize=material_factor_dict[material_type],
            doc="Material factor",
        )

        alpha_dict = {
            CompressorType.Centrifugal: {1: 7.58, 2: 0.8},
            CompressorType.Reciprocating: {1: 7.9661, 2: 0.8},
            CompressorType.Screw: {1: 8.1238, 2: 0.7243},
        }
        alpha = alpha_dict[compressor_type]

        # Purchase cost rule
        @blk.Constraint()
        def base_cost_per_unit_eq(blk):
            return blk.base_cost_per_unit == (
                pyo.exp(alpha[1] + alpha[2] * pyo.log(work_hp / pyo.units.hp))
                * pyo.units.USD_CE500
            )

        @blk.Expression(doc="Base cost for all units installed")
        def base_cost(blk):
            return blk.base_cost_per_unit * blk.number_of_units

        @blk.Constraint()
        def capital_cost_constraint(blk):
            return blk.capital_cost == (
                blk.drive_factor * blk.material_factor * blk.base_cost
            )

    def cost_fan(
        blk,
        fan_type=FanType.CentrifugalBackward,
        fan_head_factor=1.45,
        material_type=FanMaterial.StainlessSteel,
        integer=True,
    ):
        """
        Generic costing method for fans.

        Args:
            fan_type: FanType Enum indicating type of type of equipment,
                default = FanType.CentrifugalBackward.
            fan_head_factor: (float) fan head factor (default=1.45).
            material_type: FanMaterial Enum indicating material of
                construction, default = FanMaterial.StainlessSteel.
            integer: whether the number of units should be constrained to be
                an integer or not (default = True).
        """
        # Confirm that unit is a turbine
        if not blk.unit_model.config.compressor:
            raise TypeError(
                "cost_fan method is only appropriate for "
                "pressure changers with the compressor argument "
                "equal to True."
            )

        # Build generic costing variables
        _make_common_vars(blk, integer)

        t0 = blk.unit_model.flowsheet().time.first()
        Q = blk.unit_model.control_volume.properties_in[t0].flow_vol

        Qcfm = pyo.units.convert(
            Q / blk.number_of_units, to_units=pyo.units.foot**3 / pyo.units.minute
        )

        # Fan cost correlation
        alpha_dict = {
            FanType.CentrifugalBackward: {1: 11.0757, 2: 1.12906, 3: 0.08860},
            FanType.CentrifugalStraight: {1: 12.1678, 2: 1.31363, 3: 0.09974},
            FanType.VaneAxial: {1: 9.5229, 2: 0.97566, 3: 0.08532},
            FanType.TubeAxial: {1: 6.12905, 2: 0.40254, 3: 0.05787},
        }
        alpha = alpha_dict[fan_type]

        blk.head_factor = pyo.Param(
            initialize=fan_head_factor, mutable=True, doc="Fan head factor"
        )

        material_factor_dict = {
            FanMaterial.CarbonSteel: 1.0,
            FanMaterial.Fiberglass: 1.8,
            FanMaterial.StainlessSteel: 2.5,
            FanMaterial.NickelAlloy: 5.0,
        }
        blk.material_factor = pyo.Param(
            initialize=material_factor_dict[material_type],
            mutable=True,
            doc="Material factor",
        )

        # Base cost
        @blk.Constraint()
        def base_cost_per_unit_eq(blk):
            return blk.base_cost_per_unit == (
                pyo.exp(
                    alpha[1]
                    - alpha[2] * pyo.log(Qcfm / (pyo.units.foot**3 / pyo.units.minute))
                    + alpha[3]
                    * pyo.log(Qcfm / (pyo.units.foot**3 / pyo.units.minute)) ** 2
                )
                * pyo.units.USD_CE500
            )

        @blk.Expression(doc="Base cost for all units installed")
        def base_cost(blk):
            return blk.base_cost_per_unit * blk.number_of_units

        @blk.Constraint()
        def capital_cost_constraint(blk):
            return blk.capital_cost == (
                blk.material_factor * blk.head_factor * blk.base_cost
            )

    def cost_blower(
        blk,
        blower_type=BlowerType.Centrifugal,
        material_type=BlowerMaterial.StainlessSteel,
        integer=True,
    ):
        """
        Generic costing method for blowers.

        Args:
            blower_type: BlowerType Enum indicating type of type of equipment,
                default = BlowerType.Centrifugal.
            material_type: BlowerMaterial Enum indicating material of
                construction, default = BlowerMaterial.StainlessSteel.
            integer: whether the number of units should be constrained to be
                an integer or not (default = True).
        """
        # Confirm that unit is a turbine
        if not blk.unit_model.config.compressor:
            raise TypeError(
                "cost_blower method is only appropriate for "
                "pressure changers with the compressor argument "
                "equal to True."
            )

        # Build generic costing variables
        _make_common_vars(blk, integer)

        t0 = blk.unit_model.flowsheet().time.first()
        work = blk.unit_model.work_mechanical[t0]

        work_hp = pyo.units.convert(work / blk.number_of_units, to_units=pyo.units.hp)

        # Fan cost correlation
        alpha_dict = {
            BlowerType.Centrifugal: {1: 6.8929, 2: 0.7900, 3: 0.0},
            BlowerType.Rotary: {1: 7.59176, 2: 0.7932, 3: 0.012900},
        }
        alpha = alpha_dict[blower_type]

        material_factor_dict = {
            BlowerMaterial.CarbonSteel: 1.0,
            BlowerMaterial.Aluminum: 0.60,
            BlowerMaterial.Fiberglass: 1.8,
            BlowerMaterial.StainlessSteel: 2.5,
            BlowerMaterial.NickelAlloy: 5.0,
        }
        blk.material_factor = pyo.Param(
            initialize=material_factor_dict[material_type],
            mutable=True,
            doc="Material factor",
        )

        # Base cost
        @blk.Constraint()
        def base_cost_per_unit_eq(blk):
            return blk.base_cost_per_unit == (
                pyo.exp(
                    alpha[1]
                    + alpha[2] * pyo.log(work_hp / pyo.units.hp)
                    - alpha[3] * pyo.log(work_hp / pyo.units.hp) ** 2
                )
                * pyo.units.USD_CE500
            )

        @blk.Expression(doc="Base cost for all units installed")
        def base_cost(blk):
            return blk.base_cost_per_unit * blk.number_of_units

        @blk.Constraint()
        def capital_cost_constraint(blk):
            return blk.capital_cost == blk.material_factor * blk.base_cost

    def cost_turbine(blk, integer=True):
        """
        Generic costing method for turbines.

        Args:
            integer: whether the number of units should be constrained to be
                an integer or not (default = True).
        """
        # Confirm that unit is a turbine
        if blk.unit_model.config.compressor:
            raise TypeError(
                "cost_turbine method is only appropriate for "
                "pressure changers with the compressor argument "
                "equal to False."
            )

        # Build costing variables
        blk.capital_cost = pyo.Var(
            initialize=1e4,
            domain=pyo.NonNegativeReals,
            bounds=(0, None),
            units=pyo.units.USD_CE500,
            doc="Capital cost of all units",
        )

        if integer is True:
            domain = pyo.Integers
        else:
            domain = pyo.NonNegativeReals
        blk.number_of_units = pyo.Var(
            initialize=1,
            domain=domain,
            bounds=(1, 100),
            doc="Number of units to install.",
        )
        blk.number_of_units.fix(1)

        work = -blk.unit_model.work_mechanical[blk.unit_model.flowsheet().time.first()]

        work_hp = pyo.units.convert(work / blk.number_of_units, to_units=pyo.units.hp)

        @blk.Constraint()
        def capital_cost_constraint(blk):
            return blk.capital_cost == (
                blk.number_of_units
                * 530
                * (work_hp / pyo.units.hp) ** 0.81
                * pyo.units.USD_CE500
            )

    def cost_pump(
        blk,
        pump_type=PumpType.Centrifugal,
        material_type=PumpMaterial.StainlessSteel,
        pump_type_factor=1.4,
        motor_type=PumpMotorType.Open,
        integer=True,
    ):
        """
        Generic costing method for pumps.

        Args:
            pump_type: PumpType Enum indicating type of type of equipment,
                default = PumpType.Centrifugal.
            material_type: PumpMaterial Enum indicating material of
                construction, default = PumpMaterial.StainlessSteel.
                Material type is tied to PumpType.
            pump_type_factor: empirical factor for centrigual pumps based on
                table in source. Valid values are [1.1, 1.2,  1.3, 1.4 (default),
                2.1, 2.2].
            motor_type: PumpMotorType Enum indicating type of type of motor
                to be used, default = PumpMotorType.Open.
            integer: whether the number of units should be constrained to be
                an integer or not (default = True).
        """
        # Confirm that unit is a pump/compressor
        if not blk.unit_model.config.compressor:
            raise TypeError(
                "cost_pump method is only appropriate for "
                "pressure changers with the compressor argument "
                "equal to True."
            )

        # Validate argument combinations
        if pump_type == PumpType.Reciprocating:
            if material_type not in [
                PumpMaterial.DuctileIron,
                PumpMaterial.NiAlBronze,
                PumpMaterial.CarbonSteel,
                PumpMaterial.StainlessSteel,
            ]:
                raise ConfigurationError(
                    f"{blk.name} invalid combination of arguments. If "
                    "pump_type == PumpType.Reciprocating then material_type "
                    "must be one of PumpMaterial.ductileIron, "
                    "PumpMaterial.NiAlBronze, PumpMaterial.CarbonSteel, or "
                    "PumpMaterial.StainlessSteel."
                )
        else:
            if material_type in [PumpMaterial.NiAlBronze, PumpMaterial.CarbonSteel]:
                raise ConfigurationError(
                    f"{blk.name} invalid combination of arguments. "
                    "PumpMaterial.NiAlBronze and "
                    "PumpMaterial.CarbonSteel material_type"
                    " are only valid for pump_type == PumpType.Reciprocating."
                )

        # Add common variables
        blk.capital_cost = pyo.Var(
            initialize=1e4,
            domain=pyo.NonNegativeReals,
            bounds=(0, None),
            units=pyo.units.USD_CE500,
            doc="Capital cost of all units",
        )

        if integer is True:
            domain = pyo.Integers
        else:
            domain = pyo.NonNegativeReals
        blk.number_of_units = pyo.Var(
            initialize=1,
            domain=domain,
            bounds=(1, 100),
            doc="Number of units to install.",
        )
        blk.number_of_units.fix(1)

        t0 = blk.unit_model.flowsheet().time.first()

        work = pyo.units.convert(
            blk.unit_model.work_mechanical[t0] / blk.number_of_units,
            to_units=pyo.units.hp,
        )

        # Pressure units required in lbf/ft^2
        deltaP = pyo.units.convert(
            blk.unit_model.deltaP[t0],
            to_units=pyo.units.psi * pyo.units.inch**2 / pyo.units.foot**2,
        )

        # Mass density units required in lb/ft^3
        dens_mass = pyo.units.convert(
            blk.unit_model.control_volume.properties_in[t0].dens_mass,
            to_units=pyo.units.pound / pyo.units.foot**3,
        )

        # Volumetric flow units required in gpm
        Q = pyo.units.convert(
            blk.unit_model.control_volume.properties_in[t0].flow_vol
            / blk.number_of_units,
            to_units=pyo.units.gallon / pyo.units.minute,
        )

        # Calculate pump head
        blk.pump_head = pyo.Var(
            initialize=10,
            domain=pyo.NonNegativeReals,
            doc="Pump Head in feet of fluid flowing (Pressure rise/density)",
            units=pyo.units.pound_force * pyo.units.foot / pyo.units.pound,
        )

        @blk.Constraint()
        def pump_head_eq(blk):
            return blk.pump_head == deltaP / dens_mass

        # Pump size factor: S = Q(H)**0.5
        # Q = is the flow rate through the pump in gallons per minute
        # H = pump head in feet of flowing (pressure rise/liquid density)
        blk.size_factor = pyo.Var(
            initialize=10000,
            domain=pyo.NonNegativeReals,
            doc="Pump size factor, f(Q,pump_head)",
        )

        @blk.Constraint()
        def size_factor_eq(blk):
            return blk.size_factor == (
                Q
                * pyo.units.minute
                / pyo.units.gallon
                * (
                    blk.pump_head
                    * pyo.units.pound
                    / pyo.units.pound_force
                    / pyo.units.foot
                )
                ** 0.5
            )

        # Material factor
        if pump_type != PumpType.Reciprocating:
            material_factor_dict = {
                PumpMaterial.CastIron: 1.00,
                PumpMaterial.DuctileIron: 1.15,
                PumpMaterial.CastSteel: 1.35,
                PumpMaterial.Bronze: 1.90,
                PumpMaterial.StainlessSteel: 2.00,
                PumpMaterial.HastelloyC: 2.95,
                PumpMaterial.Monel: 3.30,
                PumpMaterial.Nickel: 3.50,
                PumpMaterial.Titanium: 9.70,
            }
        else:
            material_factor_dict = {
                PumpMaterial.DuctileIron: 1.00,
                PumpMaterial.StainlessSteel: 2.20,
                PumpMaterial.NiAlBronze: 1.15,
                PumpMaterial.CarbonSteel: 1.50,
            }
        blk.material_factor = pyo.Param(
            initialize=material_factor_dict[material_type],
            mutable=True,
            doc="Construction material correction factor",
        )

        # Pump type factor
        if pump_type == PumpType.Centrifugal:
            pump_type_factor_dict = {
                1.1: 1.00,
                1.2: 1.50,
                1.3: 1.70,
                1.4: 2.00,
                2.1: 2.70,
                2.2: 8.90,
            }
            try:
                ptf = pump_type_factor_dict[pump_type_factor]
            except KeyError:
                raise ConfigurationError(
                    f"{blk.name} invalid value for pump_type_factor argument: "
                    f"{pump_type_factor}. Value must be one of [1.1, 1.2, 1.3,"
                    " 1.4, 2.1, 2.2]"
                )
        else:
            ptf = 1

        blk.FT = pyo.Param(mutable=True, initialize=ptf, doc="Pump type factor")

        # Base pump cost per unit
        blk.base_pump_cost_per_unit = pyo.Var(
            initialize=1e5,
            domain=pyo.NonNegativeReals,
            units=pyo.units.USD_CE394,
            doc="Base cost of pump (less motor) per unit",
        )

        @blk.Constraint()
        def base_pump_cost_per_unit_eq(blk):
            if pump_type == PumpType.Centrifugal:
                bpc = pyo.exp(
                    9.7171
                    - 0.6019 * pyo.log(blk.size_factor)
                    + 0.0519 * pyo.log(blk.size_factor) ** 2
                )
            elif pump_type == PumpType.ExternalGear:
                bpc = pyo.exp(
                    7.6964
                    + 0.1986 * pyo.log(Q / (pyo.units.gallon / pyo.units.minute))
                    + 0.0291 * pyo.log(Q / (pyo.units.gallon / pyo.units.minute)) ** 2
                )
            elif pump_type == PumpType.Reciprocating:
                bpc = pyo.exp(
                    7.8103
                    + 0.26986 * pyo.log(work / pyo.units.hp)
                    + 0.06718 * pyo.log(work / pyo.units.hp) ** 2
                )
            return blk.base_pump_cost_per_unit == bpc * pyo.units.USD_CE394

        @blk.Expression(doc="Base cost for all pumps (less motors)")
        def base_pump_cost(blk):
            return blk.base_pump_cost_per_unit * blk.number_of_units

        blk.pump_capital_cost = pyo.Var(
            initialize=100000,
            domain=pyo.NonNegativeReals,
            units=pyo.units.USD_CE500,
            doc="Capital cost of pumps (less motors)",
        )

        @blk.Constraint()
        def pump_capital_cost_eq(blk):
            return blk.pump_capital_cost == pyo.units.convert(
                blk.FT * blk.material_factor * blk.base_pump_cost,
                to_units=pyo.units.USD_CE500,
            )

        # Motor Costs
        pump_motor_type_dict = {
            PumpMotorType.Open: 1,
            PumpMotorType.Enclosed: 1.4,
            PumpMotorType.ExplosionProof: 1.8,
        }
        blk.motor_FT = pyo.Param(
            mutable=True,
            initialize=pump_motor_type_dict[motor_type],
            doc="Motor type factor",
        )

        # Efficiency of the electric motor
        eta_m = (
            0.80
            + 0.0319 * pyo.log(work / pyo.units.hp)
            - 0.00182 * pyo.log(work / pyo.units.hp) ** 2
        )
        work_motor = work / eta_m

        blk.base_motor_cost_per_unit = pyo.Var(
            initialize=10000,
            domain=pyo.NonNegativeReals,
            units=pyo.units.USD_CE394,
            doc="Motor base purchase cost per unit",
        )

        @blk.Constraint()
        def base_motor_cost_eq(blk):
            return blk.base_motor_cost_per_unit == (
                pyo.exp(
                    5.8259
                    + 0.13141 * pyo.log(work_motor / pyo.units.hp)
                    + 0.053255 * pyo.log(work_motor / pyo.units.hp) ** 2
                    + 0.028628 * pyo.log(work_motor / pyo.units.hp) ** 3
                    - 0.0035549 * pyo.log(work_motor / pyo.units.hp) ** 4
                )
                * pyo.units.USD_CE394
            )

        @blk.Expression(doc="Base cost for all motors")
        def motor_base_cost(blk):
            return blk.base_motor_cost_per_unit * blk.number_of_units

        blk.motor_capital_cost = pyo.Var(
            initialize=100000,
            domain=pyo.NonNegativeReals,
            units=pyo.units.USD_CE500,
            doc="Capital cost of all motors",
        )

        @blk.Constraint()
        def motor_capital_cost_eq(blk):
            return blk.motor_capital_cost == pyo.units.convert(
                blk.motor_FT * blk.motor_base_cost, to_units=pyo.units.USD_CE500
            )

        # Total capital cost (pump + electrical motor)
        @blk.Constraint()
        def capital_cost_constraint(blk):
            return blk.capital_cost == (blk.pump_capital_cost + blk.motor_capital_cost)

    def cost_pressure_changer(blk, mover_type="compressor", **kwargs):
        """
        Gateway method for costing pressure changers. This method attempts
        to determine the type of pressure changer from the compressor and
        thermodynamic assumption config arguments and then calls th
        appropriate sub-method. As such, keyword arguments to this method
        will depend on the sub-method called. Users are generally encouraged
        to call the specific sub-methods directly as required.

        Args:
            mover_type: optional argument to indicate type of pressure changer.
        """
        if not blk.unit_model.config.compressor or mover_type == "turbine":
            # Unit is a turbine
            SSLWCostingData.cost_turbine(blk, **kwargs)
        elif mover_type == "compressor":
            # Unit is a pump
            SSLWCostingData.cost_compressor(blk, **kwargs)
        elif (
            mover_type == "pump"
            or blk.unit_model.config.thermodynamic_assumption
            == ThermodynamicAssumption.pump
        ):
            # Unit is a pump
            SSLWCostingData.cost_pump(blk, **kwargs)
        elif mover_type == "blower":
            # Unit is a pump
            SSLWCostingData.cost_pump(blk, **kwargs)
        elif mover_type == "fan":
            # Unit is a pump
            SSLWCostingData.cost_fan(blk, **kwargs)
        else:
            raise ConfigurationError(
                f"{blk.name} - unrecognized value for mover_type argument: "
                f"{mover_type}."
            )

    # -------------------------------------------------------------------------
    # Map costing methods to unit model classes
    # Here we can provide a dict mapping unit model classes to costing methods
    # Even better, this is inheritance aware so e.g. Pump will first look for a
    # method assigned to Pump, and then fall back to PressureChanger
    unit_mapping = {
        CSTR: cost_vertical_vessel,
        Compressor: cost_compressor,
        Flash: cost_vertical_vessel,
        Heater: cost_fired_heater,
        HeatExchanger: cost_heat_exchanger,
        HeatExchangerNTU: cost_heat_exchanger,
        PFR: cost_horizontal_vessel,
        PressureChanger: cost_pressure_changer,
        Pump: cost_pump,
        StoichiometricReactor: cost_horizontal_vessel,
        Turbine: cost_turbine,
    }


# -----------------------------------------------------------------------------
def _make_common_vars(blk, integer=True):
    # Build generic costing variables (most costing models need these vars)
    blk.base_cost_per_unit = pyo.Var(
        initialize=1e5,
        domain=pyo.NonNegativeReals,
        units=pyo.units.USD_CE500,
        doc="Base cost per unit",
    )

    blk.capital_cost = pyo.Var(
        initialize=1e4,
        domain=pyo.NonNegativeReals,
        bounds=(0, None),
        units=pyo.units.USD_CE500,
        doc="Capital cost of all units",
    )

    if integer is True:
        domain = pyo.Integers
    else:
        domain = pyo.NonNegativeReals
    blk.number_of_units = pyo.Var(
        initialize=1, domain=domain, bounds=(1, 100), doc="Number of units to install."
    )
    blk.number_of_units.fix(1)
